# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "10/01/2018"


import logging
from ..utils import WidgetLongProcessing
from silx.gui import qt
from Orange.widgets import gui
from orangecontrib.tomwer.orange.managedprocess import SuperviseOW
from Orange.widgets.settings import Setting
from Orange.widgets.widget import Input, Output
from tomwer.core.process.reconstruction.darkref.darkrefs import DarkRefs
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.reconstruction.darkref.darkrefcopywidget import DarkRefAndCopyWidget
from tomwer.synctools.ftseries import QReconsParams
from tomwer.web.client import OWClient
from processview.core.manager import DatasetState
from processview.core.manager import ProcessManager
from tomwer.utils import docstring
import copy

logger = logging.getLogger(__name__)


class DarkRefAndCopyOW(SuperviseOW, OWClient, WidgetLongProcessing):
    """
    A simple widget managing the copy of an incoming folder to an other one

    :param parent: the parent widget
    """

    # note of this widget should be the one registred on the documentation
    name = "dark and flat field construction"
    id = "orange.widgets.tomwer.darkrefs"
    description = "This widget will generate dark refs for a received scan "
    icon = "icons/darkref.svg"
    priority = 25
    category = "esrfWidgets"
    keywords = ["tomography", "dark", "darks", "ref", "refs"]

    want_main_area = True
    resizing_enabled = True
    compress_signal = False

    _rpSetting = Setting(dict())
    """Setting to load and save DarkRefAndCopyWidget settings"""

    sigScanReady = qt.Signal(TomwerScanBase)
    """Signal emitted when a scan is ready"""

    assert len(DarkRefs.inputs) == 1

    class Inputs:
        data_in = Input(
            name=DarkRefs.inputs[0].name,
            type=DarkRefs.inputs[0].type,
            doc=DarkRefs.inputs[0].doc,
        )

    assert len(DarkRefs.outputs) == 1

    class Outputs:
        data_out = Output(
            name=DarkRefs.outputs[0].name,
            type=DarkRefs.outputs[0].type,
            doc=DarkRefs.outputs[0].doc,
        )

    def __init__(self, parent=None, reconsparams=None):
        """

        :param QReconsParams reconsparams: reconstruction parameters
        """
        SuperviseOW.__init__(self, parent)
        OWClient.__init__(self)
        recons_params = reconsparams or QReconsParams()
        if self._rpSetting != dict():
            try:
                recons_params.dkrf.load_from_dict(self._rpSetting)
            except:
                logger.warning("fail to load reconstruction settings")

        self.widget = DarkRefAndCopyWidget(
            parent=self, reconsparams=recons_params, process_id=self.process_id
        )
        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self._layout.addWidget(self.widget)
        self.setForceSync = self.widget.setForceSync

        # expose API
        self.hasRefStored = self.widget.hasFlatStored
        self.setModeAuto = self.widget.set_mode_auto
        self.setRefsFromScan = self.widget.setRefsFromScan
        self.setCopyActive = self.widget.setCopyActive
        self.hasDarkStored = self.widget.hasDarkStored
        self.hasFlatStored = self.widget.hasFlatStored

        # connect signal / slot
        self.widget.sigProcessingStart.connect(self._startProcessing)
        self.widget.sigProcessingEnd.connect(self._endProcessing)
        self.widget.sigScanReady.connect(self.signalReady)
        self.widget.recons_params.sigChanged.connect(self._updateSettingsVals)
        self.widget.sigModeAutoChanged.connect(self._updateSettingsVals)
        self.widget.sigCopyActivationChanged.connect(self._updateSettingsVals)

        # load some other copy parameters
        if self._rpSetting != dict():
            try:
                if "activate" in self._rpSetting:
                    self.widget.setCopyActive(self._rpSetting["activate"])
                    del self._rpSetting["activate"]
                if "auto" in self._rpSetting:
                    self.widget.setModeAuto(self._rpSetting["auto"])
                    del self._rpSetting["auto"]
            except:
                logger.warning("fail to load reconstruction settings")

        self.setCaption(self.windowTitle())

    @Inputs.data_in
    def process(self, scanID):
        if scanID is None:
            return
        assert isinstance(scanID, TomwerScanBase)
        ProcessManager().notify_dataset_state(
            dataset=scanID, process=self, state=DatasetState.PENDING
        )
        return self.widget.process(copy.copy(scanID))

    @docstring(SuperviseOW)
    def reprocess(self, dataset):
        self.process(dataset)

    def signalReady(self, scanID):
        assert isinstance(scanID, TomwerScanBase)
        self.Outputs.data_out.send(scanID)
        self.sigScanReady.emit(scanID)

    def _updateSettingsVals(self):
        self._rpSetting = self.widget.recons_params.to_dict()
        self._rpSetting["auto"] = self.widget.isOnModeAuto()
        self._rpSetting["activate"] = self.widget.isCopyActive()

    @property
    def recons_params(self):
        return self.widget.recons_params

    def close(self):
        self.widget.close()
        super(DarkRefAndCopyOW, self).close()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "18/02/2018"


from tomwer.gui import icons as tomwericons
from silx.gui import qt


class HistoryAction(qt.QAction):
    """
    Action displaying the history of finished scans
    """

    def __init__(self, parent):
        icon = tomwericons.getQIcon("history")
        qt.QAction.__init__(self, icon, "history", parent)
        self.setCheckable(True)


class ConfigurationAction(qt.QAction):
    """
    Action to show the configuration dialog
    """

    def __init__(self, parent):
        icon = tomwericons.getQIcon("parameters")
        qt.QAction.__init__(self, icon, "configuration", parent)
        self.setCheckable(True)


class ObservationAction(qt.QAction):
    """
    Action to show the observation dialog
    """

    def __init__(self, parent):
        icon = tomwericons.getQIcon("loop")
        qt.QAction.__init__(self, icon, "observations", parent)
        self.setCheckable(True)


class ControlAction(qt.QAction):
    """
    Action to control the datawatcher (see status and select folder)
    """

    def __init__(self, parent):
        icon = tomwericons.getQIcon("health")
        qt.QAction.__init__(self, icon, "control", parent)
        self.setCheckable(True)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "24/01/2017"

import logging
import os
import shutil
import tempfile
import time
import unittest

from silx.gui import qt
from orangecontrib.tomwer.widgets.reconstruction.DarkRefAndCopyOW import (
    DarkRefAndCopyOW,
)
from tomwer.core.process.reconstruction.darkref.params import Method as DkrfMethod
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.gui.reconstruction.ftserie import FtserieWidget
from tomwer.gui.qtapplicationmanager import QApplicationManager
from tomwer.gui.reconstruction.darkref.darkrefcopywidget import DarkRefAndCopyWidget
from tomwer.synctools.ftseries import QReconsParams
from tomwer.test.utils import UtilsTest
from silx.gui.utils.testutils import TestCaseQt
from tomwer.core.utils.scanutils import MockHDF5, MockEDF
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.process.reconstruction.darkref.darkrefs import DarkRefs
from tomwer.core.process.reconstruction.darkref.settings import (
    DARKHST_PREFIX,
    REFHST_PREFIX,
)

_qapp = QApplicationManager()

logging.disable(logging.INFO)


class TestEDFDarkRefWidget(TestCaseQt):
    """class testing the DarkRefWidget"""

    def setUp(self):
        TestCaseQt.setUp(self)
        self._recons_params = QReconsParams()
        self.widget = DarkRefAndCopyOW(
            parent=None, reconsparams=self._recons_params.dkrf
        )

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        TestCaseQt.tearDown(self)

    def testSyncRead(self):
        """Make sure any modification on the self._reconsParams is
        applied on the GUI"""
        rp = self._recons_params
        self.assertTrue(rp.dkrf["REFSRMV"] is False)
        self.assertFalse(
            self.widget.widget.mainWidget.tabGeneral._rmOptionCB.isChecked()
        )
        rp.dkrf["REFSRMV"] = True
        self.assertTrue(
            self.widget.widget.mainWidget.tabGeneral._rmOptionCB.isChecked()
        )

        pattern = self.widget.widget.mainWidget.tabExpert._refLE.text()
        newText = "popo.*"
        assert pattern != newText
        rp.dkrf["RFFILE"] = newText
        self.assertTrue(
            self.widget.widget.mainWidget.tabExpert._refLE.text() == newText
        )

    def testSyncWrite(self):
        """Test that if we edit through the :class:`DarkRefWidget` then the
        modification are fall back into the self._reconsParams"""
        rp = self._recons_params

        # test patterns
        pattern = self.widget.widget.mainWidget.tabExpert._refLE.text()
        newText = "popo.*"
        assert pattern != newText
        self.widget.widget.mainWidget.tabExpert._refLE.setText(newText)
        self.widget.widget.mainWidget.tabExpert._refLE.editingFinished.emit()
        qt.QApplication.instance().processEvents()
        self.assertTrue(rp.dkrf["RFFILE"] == newText)
        self.widget.widget.mainWidget.tabExpert._darkLE.setText(newText)
        self.widget.widget.mainWidget.tabExpert._darkLE.editingFinished.emit()
        qt.QApplication.instance().processEvents()
        self.assertTrue(rp.dkrf["DKFILE"] == newText)

        # test calc mode
        self.widget.widget.mainWidget.tabGeneral._darkWCB.setMode(DkrfMethod.none)
        self.widget.widget.mainWidget.tabGeneral._refWCB.setMode(DkrfMethod.median)
        self.assertTrue(rp.dkrf["DARKCAL"] == DkrfMethod.none)
        self.assertTrue(rp.dkrf["REFSCAL"] == DkrfMethod.median)

        # test options
        cuRm = self.widget.widget.mainWidget.tabGeneral._rmOptionCB.isChecked()
        self.widget.widget.mainWidget.tabGeneral._rmOptionCB.setChecked(not cuRm)
        self.assertTrue(rp.dkrf["REFSRMV"] == (not cuRm))
        self.assertTrue(rp.dkrf["DARKRMV"] == (not cuRm))

        cuSkip = self.widget.widget.mainWidget.tabGeneral._skipOptionCB.isChecked()
        self.widget.widget.mainWidget.tabGeneral._skipOptionCB.setChecked(not cuSkip)
        # warning : here value of skip and overwrite are of course inverse
        self.assertTrue(rp.dkrf["DARKOVE"] == cuSkip)
        self.assertTrue(rp.dkrf["REFSOVE"] == cuSkip)

    def testBehaviorWithFtserie(self):
        """
        Make sure modification on ftserie 'Dark and flat field' tab are
        correctly take into account in this case they are working on the same
        QReconsParams object
        """
        ftserie = FtserieWidget(recons_params=self._recons_params)

        cuRm = ftserie.getReconsParamSetEditor()._dkRefWidget._qcbRmRef.isChecked()
        cuSkip = ftserie.getReconsParamSetEditor()._dkRefWidget._qcbSkipRef.isChecked()
        calcDK = ftserie.getReconsParamSetEditor()._dkRefWidget._qcbDKMode.getMode()
        calcRef = ftserie.getReconsParamSetEditor()._dkRefWidget._qcbRefMode.getMode()
        patternRef = (
            ftserie.getReconsParamSetEditor()._dkRefWidget._qleRefsPattern.text()
        )
        patternDK = ftserie.getReconsParamSetEditor()._dkRefWidget._qleDKPattern.text()

        # make sure initial status are the same between ftserie and darkref
        dkrfOWMainWidget = self.widget.widget.mainWidget
        self.assertTrue(dkrfOWMainWidget.tabGeneral._rmOptionCB.isChecked() == cuRm)
        self.assertTrue(dkrfOWMainWidget.tabGeneral._skipOptionCB.isChecked() == cuSkip)
        self.assertTrue(dkrfOWMainWidget.tabGeneral._darkWCB.getMode() == calcDK)
        self.assertTrue(dkrfOWMainWidget.tabGeneral._refWCB.getMode() == calcRef)
        self.assertTrue(dkrfOWMainWidget.tabExpert._refLE.text() == patternRef)
        self.assertTrue(dkrfOWMainWidget.tabExpert._darkLE.text() == patternDK)

        # change parameters
        ftserie.getReconsParamSetEditor()._dkRefWidget._qcbRmRef.setChecked(not cuRm)
        ftserie.getReconsParamSetEditor()._dkRefWidget._qcbSkipRef.setChecked(
            not cuSkip
        )
        ftserie.getReconsParamSetEditor()._dkRefWidget._qcbDKMode.setMode("none")
        ftserie.getReconsParamSetEditor()._dkRefWidget._qcbRefMode.setMode("average")
        text_ref_pattern = "toto.*"
        text_dk_pattern = "[maestro]?."
        ftserie.getReconsParamSetEditor()._dkRefWidget._qleDKPattern.setText(
            text_dk_pattern
        )
        ftserie.getReconsParamSetEditor()._dkRefWidget._qleDKPattern.editingFinished.emit()
        qt.QApplication.instance().processEvents()
        ftserie.getReconsParamSetEditor()._dkRefWidget._qleRefsPattern.setText(
            text_ref_pattern
        )
        ftserie.getReconsParamSetEditor()._dkRefWidget._qleRefsPattern.editingFinished.emit()
        qt.QApplication.instance().processEvents()

        # check modification are well take into account
        self.assertTrue(
            dkrfOWMainWidget.tabGeneral._rmOptionCB.isChecked() == (not cuRm)
        )
        self.assertTrue(self._recons_params.dkrf["DARKRMV"] == (not cuRm))
        self.assertTrue(
            dkrfOWMainWidget.tabGeneral._skipOptionCB.isChecked() == (not cuSkip)
        )
        self.assertTrue(self._recons_params.dkrf["DARKOVE"] == cuSkip)
        self.assertTrue(
            dkrfOWMainWidget.tabGeneral._darkWCB.getMode() == DkrfMethod.none
        )
        self.assertTrue(self._recons_params.dkrf["DARKCAL"] == DkrfMethod.none)
        self.assertTrue(
            dkrfOWMainWidget.tabGeneral._refWCB.getMode() == DkrfMethod.average
        )
        self.assertTrue(self._recons_params.dkrf["REFSCAL"] == DkrfMethod.average)
        self.assertTrue(self._recons_params.dkrf["DKFILE"] == text_dk_pattern)
        self.assertTrue(self._recons_params.dkrf["RFFILE"] == text_ref_pattern)
        self.assertTrue(dkrfOWMainWidget.tabExpert._refLE.text() == text_ref_pattern)
        self.assertTrue(dkrfOWMainWidget.tabExpert._darkLE.text() == text_dk_pattern)


@unittest.skipIf(
    UtilsTest.getInternalTestDir("testslicesNemoz6x") is None, "No extra datatset"
)
class TestID16TestCase(TestCaseQt):
    """
    class testing the process of the dark ref widget in the case of ID16
    """

    def setUp(self):
        TestCaseQt.setUp(self)
        datasetDir = UtilsTest.getInternalTestDir("testslicesNemoz6x")
        self._tmpDir = tempfile.mkdtemp()
        self.datasets = []
        for subFolder in (
            "testslicesNemoz61_1_",
            "testslicesNemoz62_1_",
            "testslicesNemoz63_1_",
            "testslicesNemoz64_1_",
            "testslicesNemoz65_1_",
        ):
            shutil.copytree(
                os.path.join(datasetDir, subFolder),
                os.path.join(self._tmpDir, subFolder),
            )
            self.datasets.append(os.path.join(self._tmpDir, subFolder))

        self.recons_params = QReconsParams()
        self.recons_params.dkrf.remove_dark = True
        self.recons_params.dkrf.remove_ref = True
        self.widget = DarkRefAndCopyWidget(
            parent=None, reconsparams=self.recons_params.dkrf
        )
        self.widget.setForceSync(True)

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        shutil.rmtree(self._tmpDir)
        TestCaseQt.tearDown(self)

    def test(self):
        """make sure the behavior of dark ref is correct for id16b pipeline:
        datasets is composed of :

        - testslicesNemoz61_1_: contains original dark and ref. DarkRef should
           process those to generate median ref and dark
        - testslicesNemoz62_1_, testslicesNemoz63_1_, testslicesNemoz64_1_:
           contains no ref or dark orignals or median.
           should copy the one normalized from testslicesNemoz61_1_
        testslicesNemoz65_1_: contains dark median. Should copy ref from
            testslicesNemoz61_1_
        """
        # check behavior for testslicesNemoz61_1_.
        files = os.listdir(self.datasets[0])
        assert "darkend0000.edf" in files
        assert "ref0000_0000.edf" in files
        assert "ref0000_0050.edf" in files
        assert "ref0001_0000.edf" in files
        assert "ref0001_0050.edf" in files
        assert "refHST0000.edf" not in files
        assert "refHST0050.edf" not in files
        assert "dark.edf" not in files

        scan = ScanFactory.create_scan_object(scan_path=self.datasets[0])
        self.widget.process(scan)
        for t in range(8):
            qt.QApplication.instance().processEvents()
            time.sleep(0.5)

        files = os.listdir(scan.path)
        self.assertTrue("dark.edf" in files)
        self.assertTrue("darkend0000.edf" not in files)
        self.assertTrue("ref0000_0000.edf" not in files)
        self.assertTrue("ref0000_0050.edf" not in files)
        self.assertTrue("ref0001_0000.edf" not in files)
        self.assertTrue("ref0001_0050.edf" not in files)
        self.assertTrue("refHST0000.edf" in files)
        self.assertTrue("refHST0050.edf" in files)
        self.assertTrue(self.widget._darkRef.has_flat_stored())

        # check behavior for testslicesNemoz62_1_, testslicesNemoz63_1_,
        # testslicesNemoz64_1_.
        for dataset in self.datasets[1:4]:
            files = os.listdir(dataset)
            assert "darkend0000.edf" not in files
            assert "ref0000_0000.edf" not in files
            assert "ref0000_0050.edf" not in files
            assert "ref0001_0000.edf" not in files
            assert "ref0001_0050.edf" not in files
            assert "refHST0000.edf" not in files
            assert "refHST0050.edf" not in files
            assert "dark.edf" not in files
            scan = ScanFactory.create_scan_object(scan_path=dataset)
            self.widget.process(scan)
            for t in range(8):
                qt.QApplication.instance().processEvents()
                time.sleep(0.5)

            files = os.listdir(dataset)
            self.assertTrue("darkend0000.edf" not in files)
            self.assertTrue("ref0000_0000.edf" not in files)
            self.assertTrue("ref0000_0050.edf" not in files)
            self.assertTrue("ref0001_0000.edf" not in files)
            self.assertTrue("ref0001_0050.edf" not in files)
            self.assertTrue("refHST0000.edf" in files)
            self.assertTrue("refHST0050.edf" in files)
            self.assertTrue("dark.edf" in files)

        # check behavior for testslicesNemoz65_1_
        dataset = self.datasets[-1]
        files = os.listdir(dataset)
        assert "darkend0000.edf" not in files
        assert "ref0000_0000.edf" not in files
        assert "ref0000_0050.edf" in files
        assert "ref0001_0000.edf" not in files
        assert "ref0001_0050.edf" in files
        assert "refHST0000.edf" not in files
        assert "refHST0050.edf" not in files
        assert "dark0000.edf" not in files
        scan = ScanFactory.create_scan_object(scan_path=dataset)
        self.widget.process(scan)
        for t in range(8):
            qt.QApplication.instance().processEvents()
            time.sleep(0.5)

        files = os.listdir(dataset)
        self.assertTrue("darkend0000.edf" not in files)
        self.assertTrue("ref0000_0000.edf" not in files)
        self.assertTrue("ref0000_0050.edf" not in files)
        self.assertTrue("ref0001_0000.edf" not in files)
        self.assertTrue("ref0001_0050.edf" not in files)
        self.assertTrue("refHST0000.edf" in files)
        self.assertTrue("refHST0050.edf" in files)
        self.assertTrue("dark.edf" in files)


class TestDarkRefCopyWithEDFAndHDF5(TestCaseQt):
    """Test the DarkRefCopy orange widget behaviour"""

    def setUp(self) -> None:
        TestCaseQt.setUp(self)
        self._folder = tempfile.mkdtemp()

        # define scans to be treated
        hdf5_mock_with_refs = MockHDF5(
            scan_path=os.path.join(self._folder, "h5_with_refs"),
            n_proj=10,
            n_ini_proj=10,
            dim=20,
            create_ini_ref=True,
            create_ini_dark=True,
            create_final_ref=False,
        )
        self.hdf5_acquisition_with_refs = hdf5_mock_with_refs.scan

        hdf5_mock_without_refs = MockHDF5(
            scan_path=os.path.join(self._folder, "h5_without_refs"),
            n_proj=10,
            n_ini_proj=10,
            dim=20,
            create_ini_ref=False,
            create_ini_dark=False,
        )
        self.hdf5_acquisition_without_refs = hdf5_mock_without_refs.scan

        hdf5_mock_without_refs_incoherent_dim = MockHDF5(
            scan_path=os.path.join(self._folder, "h5_without_refs_different_dim"),
            n_proj=10,
            n_ini_proj=10,
            dim=21,
            create_ini_ref=False,
            create_ini_dark=False,
        )
        self.hdf5_acquisition_without_refs_incoherent_dim = (
            hdf5_mock_without_refs_incoherent_dim.scan
        )

        edf_mock_without_ref = MockEDF(
            scan_path=os.path.join(self._folder, "edf_without_refs"), dim=20, n_radio=20
        )
        self.edf_acquisition_without_ref = EDFTomoScan(
            scan=edf_mock_without_ref.scan_path
        )

        # processes set up
        self._recons_params = QReconsParams()
        self.widget = DarkRefAndCopyOW(
            parent=None, reconsparams=self._recons_params.dkrf
        )
        self.widget.setForceSync(True)
        self.widget.show()

    def tearDown(self) -> None:
        time.sleep(0.5)
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        shutil.rmtree(self._folder)
        TestCaseQt.tearDown(self)

    def testCopyInactive(self):
        self.widget.setCopyActive(False)

        self.widget.process(self.hdf5_acquisition_with_refs)
        self.assertTrue(os.path.exists(self.hdf5_acquisition_with_refs.process_file))

        # make sure dark has been processed for the one with ref
        self.assertEqual(
            len(
                DarkRefs.get_flats_frm_process_file(
                    self.hdf5_acquisition_with_refs.process_file
                )
            ),
            1,
        )
        self.assertEqual(
            len(
                DarkRefs.get_darks_frm_process_file(
                    self.hdf5_acquisition_with_refs.process_file
                )
            ),
            1,
        )
        # make sure no dark or flats exists for the hdf5 without refs
        for scan in (
            self.hdf5_acquisition_without_refs_incoherent_dim,
            self.hdf5_acquisition_without_refs,
        ):
            self.widget.process(scan)
            self.assertFalse(os.path.exists(scan.process_file))

        # make sure no dark or flats exists for the edf one
        self.widget.process(self.edf_acquisition_without_ref)

    def testCopyActive(self):
        self.widget.setCopyActive(True)
        self.widget.setModeAuto(True)

        self.widget.process(self.hdf5_acquisition_with_refs)
        self.assertTrue(os.path.exists(self.hdf5_acquisition_with_refs.process_file))
        # 1. make sure dark has been processed for the one with ref
        self.assertEqual(
            len(
                DarkRefs.get_flats_frm_process_file(
                    self.hdf5_acquisition_with_refs.process_file
                )
            ),
            1,
        )
        self.assertEqual(
            len(
                DarkRefs.get_darks_frm_process_file(
                    self.hdf5_acquisition_with_refs.process_file
                )
            ),
            1,
        )
        self.assertTrue(self.widget.hasDarkStored())
        self.assertTrue(self.widget.hasFlatStored())

        # 2. make sure copy has been processed for the the 'compatible hdf5'
        self.widget.process(self.hdf5_acquisition_without_refs)
        self.assertTrue(os.path.exists(self.hdf5_acquisition_without_refs.process_file))
        self.assertEqual(
            len(
                DarkRefs.get_flats_frm_process_file(
                    self.hdf5_acquisition_without_refs.process_file
                )
            ),
            2,
        )
        self.assertEqual(
            len(
                DarkRefs.get_darks_frm_process_file(
                    self.hdf5_acquisition_without_refs.process_file
                )
            ),
            1,
        )

        # 3. make sure copy has been processed for the the 'compatible edf'
        self.widget.process(self.edf_acquisition_without_ref)
        self.assertEqual(
            len(
                DarkRefs.getRefHSTFiles(
                    self.edf_acquisition_without_ref.path, prefix=REFHST_PREFIX
                )
            ),
            2,
        )
        self.assertEqual(
            len(
                DarkRefs.getDarkHSTFiles(
                    self.edf_acquisition_without_ref.path, prefix=DARKHST_PREFIX
                )
            ),
            1,
        )

        # 4. make sure process but no copy made if incompatible size
        self.widget.process(self.hdf5_acquisition_without_refs_incoherent_dim)
        self.assertFalse(
            os.path.exists(
                self.hdf5_acquisition_without_refs_incoherent_dim.process_file
            )
        )


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestEDFDarkRefWidget, TestID16TestCase, TestDarkRefCopyWithEDFAndHDF5):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# coding: utf-8
###########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

"""contain the SAAxisProcess. Half automatic center of rotation calculation
"""

__authors__ = [
    "H.Payno",
]
__license__ = "MIT"
__date__ = "10/02/2021"


from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.process.reconstruction.nabu.nabuslices import (
    interpret_tomwer_configuration,
    _get_file_basename_reconstruction,
    generate_nabu_configfile,
)
from typing import Iterable
from tomwer.core.scan.edfscan import EDFTomoScan
from ..nabu import utils
from ..nabu import settings as nabu_settings
from .params import ScoreMethod
from .params import ReconstructionMode
from .params import SAAxisParams
from tomwer.core.process.baseprocess import SingleProcess, _input_desc, _output_desc
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.utils import logconfig
from tomwer.core.process.reconstruction.nabu.utils import check_sinogram_half
from processview.core.manager import ProcessManager, DatasetState
from processview.core.superviseprocess import SuperviseProcess
from tomwer.core.process.baseprocess import BaseProcess
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomoscan.io import HDF5File
from silx.io.utils import get_data
import tomwer.version
import numpy
import subprocess
from tomwer.io.utils import format_stderr_stdout
from tomwer.core.progress import Progress
from tomwer.core.process.reconstruction.axis import AxisRP
from typing import Union
import os
import logging
import copy
import h5py
import sys

try:
    from nabu.app.local_reconstruction import LocalReconstruction
except ImportError as e:
    has_nabu = False
else:
    has_nabu = True


_logger = logging.getLogger(__name__)


DEFAULT_RECONS_FOLDER = "saaxis_results"


class _ComputedScore:
    def __init__(self, tv, std):
        self._tv = tv
        self._std = std

    @property
    def total_variation(self):
        return self._tv

    @property
    def std(self):
        return self._std

    def get(self, method: ScoreMethod):
        method = ScoreMethod.from_value(method)
        if method is ScoreMethod.TV:
            return self.total_variation
        elif method is ScoreMethod.STD:
            return self.std
        else:
            raise ValueError("{} is an unrecognized method".format(method))


def compute_score_contrast_std(data: numpy.ndarray):
    """
    Compute a contrast score by simply computing the standard deviation of
    the frame
    :param numpy.ndarray data: frame for which we should compute the score
    :return: score of the frame
    :rtype: float
    """
    if data is None:
        return None
    elif data.ndim != 2:
        raise ValueError("Data is expected to be 2D. Not {}".format(data.ndim))
    else:
        return data.std() * 100


def compute_tv_score(data: numpy.ndarray):
    """
    Compute the data score as image total variation

    :param numpy.ndarray data: frame for which we should compute the score
                                Handle
    :return: score of the frame
    :rtype: float
    """
    if data is None:
        return None
    elif data.ndim != 2:
        raise ValueError("Data is expected to be 2D. Not {}".format(data.ndim))
    else:
        # handle mask if data is a numpy.ma.array gradient might set "--"
        tv = numpy.sum(
            numpy.sqrt(
                numpy.gradient(data, axis=0) ** 2 + numpy.gradient(data, axis=1) ** 2
            )
        )
        if tv > 0:
            # adapt score to:
            #    - get growing score: the higher the score is the better the cor is.
            #      this is the 1 / tv part
            #    - look more "friendly" (10e5 part)
            return (1.0 / tv) * float(10e5)
        else:
            return tv


_METHOD_TO_FCT = {
    ScoreMethod.STD: compute_score_contrast_std,
    ScoreMethod.TV: compute_tv_score,
}


def compute_score(data: numpy.ndarray, method: ScoreMethod) -> float:
    """

    :param numpy.ndarray data: frame for which we should compute the score
    :param str method:
    :return: score of the frame
    :rtype: float
    """
    if not data.ndim == 2:
        raise ValueError("Data is expected to be 2D. Not {}".format(data.ndim))
    method = ScoreMethod.from_value(method)
    fct = _METHOD_TO_FCT.get(method, None)
    if fct is not None:
        return fct(data)
    else:
        raise ValueError("{} is not handled".format(method))


def one_slice_several_cor(scan, configuration: Union[dict, SAAxisParams]) -> tuple:
    """
    Run a slice reconstruction using nabu per Center Of Rotation (cor) provided
    Then for each compute a score (quality) of the center of rotation

    :param TomwerScanBase scan:
    :param dict configuration:
    :return: cor_reconstructions, outs, errs
             cor_reconstructions is a dictionary of cor as key and a tuple
             (url, score) as value
    :rtype: tuple
    """
    if isinstance(configuration, dict):
        configuration = SAAxisParams.from_dict(configuration)
    elif not isinstance(configuration, SAAxisParams):
        raise TypeError(
            "configuration should be a dictionary or an instance of SAAxisParams"
        )
    check_sinogram_half(scan)

    configuration.check_configuration()
    mode = ReconstructionMode.from_value(configuration.mode)
    slice_index = configuration.slice_indexes
    cors = configuration.cors
    nabu_config = configuration.nabu_params
    output_dir = configuration.output_dir
    dry_run = configuration.dry_run
    nabu_output_config = nabu_config.get("output", {})
    file_format = nabu_output_config.get("file_format", "hdf5")
    _logger.info(
        "launch reconstruction of slice {} and cors {}".format(slice_index, cors)
    )
    if mode is ReconstructionMode.VERTICAL:
        if isinstance(slice_index, str):
            if not slice_index == "middle":
                raise ValueError("slice index {} not recognized".format(slice_index))
        elif not len(slice_index) == 1:
            raise ValueError("{} mode only manage one slice".format(mode.value))
        else:
            slice_index = list(slice_index.values())[0]
        advancement = Progress("saaxis - slice {} of {}".format(slice_index, scan))

        _, cor_reconstructions, outs, errs = run_slice_reconstruction(
            scan=scan,
            slice_index=slice_index,
            cor_positions=cors,
            config=nabu_config,
            output_dir=output_dir,
            dry_run=dry_run,
            file_format=file_format,
            advancement=advancement,
        )
    else:
        raise ValueError("{} is not handled for now".format(mode))
    scores = {}

    def load_datasets():
        datasets_ = {}
        for cor, urls in cor_reconstructions.items():
            for url in urls:
                try:
                    data = get_data(url=url)
                except Exception as e:
                    _logger.error(
                        "Fail to compute a score for {}. Reason is {}"
                        "".format(url.path(), str(e))
                    )
                    datasets_[cor] = (url, None)
                else:
                    if data.ndim == 3:
                        if data.shape[0] == 1:
                            data = data.reshape(data.shape[1], data.shape[2])
                        elif data.shape[2] == 1:
                            data = data.reshape(data.shape[0], data.shape[1])
                        else:
                            raise ValueError(
                                "Data is expected to be 2D. Not {}".format(data.ndim)
                            )
                    elif data.ndim == 2:
                        pass
                    else:
                        raise ValueError(
                            "Data is expected to be 2D. Not {}".format(data.ndim)
                        )

                    datasets_[cor] = (url, data)
        return datasets_

    datasets = load_datasets()

    def get_disk_mask_radius(datasets_) -> int:
        """compute the radius to use for the mask"""
        radius = sys.maxsize
        # get min radius
        for cor, (url, data) in datasets_.items():
            assert data.ndim is 2, "data is expected to be 2D"
            min_ = numpy.array(data.shape).min()
            if radius >= min_:
                radius = min_
        return radius // 2

    def apply_roi(data, radius, url) -> numpy.array:
        """compute the square included in the circle of radius and centered
        in the middle of the data"""
        half_width = int(radius / 2 ** 0.5)
        center = numpy.array(data.shape[:]) // 2
        min_x, max_x = center[0] - half_width, center[0] + half_width
        min_y, max_y = center[1] - half_width, center[1] + half_width
        try:
            return data[min_y:max_y, min_x:max_x]
        except Exception:
            _logger.error(
                "Fail to apply roi for {}. Take the entire dataset".format(url.path())
            )
            return data

    mask_disk_radius = get_disk_mask_radius(datasets)
    for cor, (url, data) in datasets.items():
        if data is None:
            score = None
        else:
            assert data.ndim == 2
            data_roi = apply_roi(data=data, radius=mask_disk_radius, url=url)
            score = _ComputedScore(
                tv=compute_score(data=data_roi, method=ScoreMethod.TV),
                std=compute_score(data=data_roi, method=ScoreMethod.STD),
            )
        scores[cor] = (url, score)

    return scores, outs, errs


def run_slice_reconstruction(
    scan: TomwerScanBase,
    cor_positions: Iterable,
    slice_index: int,
    config: dict,
    output_dir=None,
    dry_run: bool = False,
    local: bool = True,
    file_format: str = "hdf5",
    advancement=None,
) -> tuple:
    """
    call nabu for a reconstruction on scan with the given configuration

    :param TomwerScanBase scan: scan to reconstruct
    :param tuple: cor_positions cor position to used for reconstruction
    :param dict config: configuration to run the reconstruction
    :param Union[None,str]: output dir folder. If None then this will be store
                            under the acquisition folder/saaxis_results
    :param bool dry_run: do we want to run dry
    :param bool local: do we want to run a local reconstruction
    :param advancement: optional Progress class to display advancement

    :return: success, recons_urls, outs, errs
             recons_urls is a dict with cor value as key (float) and reconstructed slice url
             as value
    :rtype: dict
    """
    nabu_configurations = interpret_tomwer_configuration(config, scan=None)
    if len(nabu_configurations) == 0:
        raise RuntimeWarning(
            "Unable to get a valid nabu configuration for " "reconstruction."
        )
    elif len(nabu_configurations) > 1:
        _logger.warning(
            "Several configuration found for nabu (you probably "
            "ask for several delta/beta value or several slices). "
            "Picking the first one."
        )
    if output_dir is None:
        output_dir = os.path.join(scan.path, DEFAULT_RECONS_FOLDER)
    if scan.process_file is not None:
        steps_file_basename, _ = os.path.splitext(scan.process_file)
        steps_file_basename = "_".join(
            ("steps_file_basename", "nabu", "sinogram", "save", "step")
        )
        steps_file_basename = steps_file_basename + ".hdf5"
        steps_file = os.path.join(output_dir, steps_file_basename)
    else:
        steps_file = ""

    configs = {}

    for i_cor, cor in enumerate(cor_positions):
        nabu_configuration = copy.deepcopy(nabu_configurations[0][0])
        nabu_configuration["pipeline"] = {
            "save_steps": "sinogram" if i_cor == 0 else "",
            "resume_from_step": "sinogram",
            "steps_file": steps_file,
        }
        # convert cor from tomwer ref to nabu ref
        if scan.dim_1 is not None:
            cor_nabu_ref = cor + scan.dim_1 // 2
        else:
            _logger.warning("enable to get image half width. Set it to 1024")
            cor_nabu_ref = cor + 1024
        # handle reconstruction section
        if "reconstruction" not in nabu_configuration:
            nabu_configuration["reconstruction"] = {}
        nabu_configuration["reconstruction"]["rotation_axis_position"] = str(
            cor_nabu_ref
        )
        # handle output section
        if "output" not in nabu_configuration:
            nabu_configuration["output"] = {}
        nabu_configuration["output"]["location"] = output_dir
        nabu_configuration["output"]["file_format"] = file_format
        # handle resources section
        if local is True:
            resources_method = "local"
        else:
            resources_method = "slurm"
        nabu_configuration["resources"] = utils.get_nabu_resources_desc(
            scan=scan, workers=1, method=resources_method
        )
        configs[cor] = nabu_configuration
    return run_nabu_one_slice_several_config(
        nabu_configs=configs,
        scan=scan,
        local=local,
        slice_index=slice_index,
        dry_run=dry_run,
        file_format=file_format,
        advancement=advancement,
    )


def run_nabu_one_slice_several_config(
    scan,
    nabu_configs,
    dry_run,
    slice_index: Union[int, str],
    local,
    file_format,
    advancement=None,
) -> tuple:
    """
    # TODO: might need something like a context or an option "keep" slice in memory

    :param scan:
    :param nabu_configs: list of nabu configurations to be run
    :param dry_run:
    :param int slice_index: slice index to reconstruct or "middle"
    :param local:
    :param advancement: optional Progress class to display advancement
    :return: tuple with (success, recons_urls (list of output urls),
             tuple of outs, tuples of errs)
    """
    if slice_index == "middle":
        if scan.dim_2 is not None:
            slice_index = scan.dim_2 // 2
        else:
            _logger.warning(
                "scan.dim_2 returns None, unable to deduce middle " "pick 1024"
            )
            slice_index = 1024
    assert isinstance(slice_index, int), "slice_index should be an int"

    def preprocess_config(config, cor: float):
        dataset_params = scan.get_nabu_dataset_info()
        if "dataset" in config:
            dataset_params.update(config["dataset"])
        config["dataset"] = dataset_params

        if local is True:
            resources_method = "local"
        else:
            resources_method = "slurm"
        config["resources"] = utils.get_nabu_resources_desc(
            scan=scan, workers=1, method=resources_method
        )
        # force overwrite results
        if "output" not in config:
            config["output"] = {}
        config["output"].update({"overwrite_results": 1})

        def treateOutputConfig(_config):
            """
            - add or overwrite some parameters of the dictionary
            - create the output directory if does not exist
            """
            pag = False
            db = None
            if "phase" in _config:
                if "method" in _config["phase"] and _config["phase"]["method"] != "":
                    pag = True
                    if "delta_beta" in _config["phase"]:
                        db = round(float(_config["phase"]["delta_beta"]))
            if "output" in config:
                _file_name = _get_file_basename_reconstruction(
                    scan=scan, slice_index=slice_index, pag=pag, db=db
                )
                _config["output"]["file_prefix"] = "cor_{}_{}".format(_file_name, cor)
                if _config["output"]["location"] not in ("", None):
                    # if user specify the location
                    if not os.path.isdir(_config["output"]["location"]):
                        os.makedirs(_config["output"]["location"])
                else:
                    # otherwise default location will be the data root level
                    _config["output"]["location"] = os.sep.join(
                        scan.path, "saaxis_results"
                    )
            if "reconstruction" not in _config:
                _config["reconstruction"] = {}
            _config["reconstruction"]["start_z"] = slice_index
            _config["reconstruction"]["end_z"] = slice_index
            return _config

        config = treateOutputConfig(config)
        # the policy is to save nabu .cfg file at the same location as the
        # force overwrite results

        cfg_folder = os.path.join(
            config["output"]["location"],
            nabu_settings.NABU_CFG_FILE_FOLDER,
        )
        if not os.path.exists(cfg_folder):
            os.makedirs(cfg_folder)

        name = (
            config["output"]["file_prefix"] + nabu_settings.NABU_CONFIG_FILE_EXTENSION
        )
        if not isinstance(scan, EDFTomoScan):
            name = "_".join((scan.entry, name))
        conf_file = os.path.join(cfg_folder, "cor_{}_{}".format(cor, name))
        return config, conf_file

    recons_urls = {}
    outs = []
    errs = []
    if advancement:
        advancement.setMaxAdvancement(len(nabu_configs))
    # TODO: add logger and extra options
    for cor, config in nabu_configs.items():
        config, conf_file = preprocess_config(config, cor)
        # add some tomwer metadata and save the configuration
        # note: for now the section is ignored by nabu but shouldn't stay that way
        with utils.TomwerInfo(config) as config_to_dump:
            generate_nabu_configfile(
                fname=conf_file, config=config_to_dump, options_level="advanced"
            )
        if dry_run:
            continue

        if slice_index is not None and dry_run is False and local:
            if not has_nabu:
                raise ImportError("Fail to import nabu")
            _logger.info(
                "run nabu slice reconstruction for %s with %s" "" % (scan.path, config)
            )

            # need to be executed in his own context
            command = " ".join(
                ("python", "-m", "nabu.resources.cli.reconstruct", conf_file)
            )
            _logger.info('call nabu from "{}"'.format(command))

            process = subprocess.Popen(
                command,
                shell=True,
                cwd=scan.path,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

            out, err = process.communicate()
            out_str = format_stderr_stdout(out, err)
            outs.append(out)
            errs.append(err)

        urls = utils.get_recons_urls(
            file_prefix=config_to_dump["output"]["file_prefix"],
            location=config_to_dump["output"]["location"],
            slice_index=None,
            scan=scan,
            file_format=file_format,
            start_z=None,
            end_z=None,
        )
        # specific treatment for cor: rename output files
        recons_urls[cor] = urls
        if advancement:
            advancement.increaseAdvancement(1)
    return (
        len(recons_urls) > 0,
        recons_urls,
        tuple(outs),
        tuple(errs),
    )


class SAAxisProcess(SingleProcess, SuperviseProcess):
    """
    Main process to launch several reconstruction of a single slice with
    several Center Of Rotation (cor) values
    """

    inputs = [
        _input_desc(
            name="data", type=TomwerScanBase, handler="pathReceived", doc="scan path"
        ),
    ]
    outputs = [_output_desc(name="data", type=TomwerScanBase, doc="scan path")]

    def __init__(self, process_id=None, dump_process=True):
        SingleProcess.__init__(self)
        SuperviseProcess.__init__(self, process_id=process_id)
        self._dry_run = False
        self._dump_process = dump_process
        self._std_outs = tuple()
        self._std_errs = tuple()

    @property
    def std_outs(self):
        return self._std_outs

    @property
    def std_errs(self):
        return self._std_errs

    def set_dry_run(self, dry_run):
        self._dry_run = dry_run

    @property
    def dry_run(self):
        return self._dry_run

    def set_configuration(self, configuration: dict) -> None:
        return self.set_properties(configuration)

    def set_properties(self, properties):
        if isinstance(properties, SAAxisParams):
            self._settings = properties.to_dict()
        elif isinstance(properties, dict):
            self._settings = properties
        else:
            raise TypeError(
                "properties should be an instance of dict or " "SAAxisParams"
            )

    @staticmethod
    def autofocus(scan):
        scores = scan.saaxis_params.scores
        score_method = scan.saaxis_params.score_method
        best_cor, best_score = None, 0
        for cor, (_, score_cls) in scores.items():
            score = score_cls.get(score_method)
            if score > best_score:
                best_cor, best_score = cor, score
        scan.saaxis_params.autofocus = best_cor
        if scan.axis_params is None:
            scan.axis_params = AxisRP()
        scan.axis_params.frame_width = scan.dim_1
        scan.axis_params.set_relative_value(best_cor)

    def process(self, scan=None):
        if scan is None:
            return None
        if isinstance(scan, TomwerScanBase):
            scan = scan
        elif isinstance(scan, dict):
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            raise ValueError(
                "input type of {}: {} is not managed" "".format(scan, type(scan))
            )
        # TODO: look and update if there is some nabu reconstruction
        # or axis information to be used back
        configuration = self.get_configuration()
        params = SAAxisParams.from_dict(configuration)
        # insure output dir is created
        if params.output_dir in (None, ""):
            params.output_dir = os.path.join(scan.path, "saaxis_results")
            if not os.path.exists(params.output_dir):
                os.makedirs(params.output_dir)
        # try to find an estimated cor
        #  from a previously computed cor
        if params.estimated_cor is None and scan.axis_params is not None:
            relative_cor = scan.axis_params.relative_cor_value
            if relative_cor is not None and numpy.issubdtype(
                type(relative_cor), numpy.number
            ):
                params.estimated_cor = relative_cor
                _logger.info(
                    "{}: set estimated cor from previously computed cor ({})".format(
                        str(scan), params.estimated_cor
                    )
                )
        #  from scan.estimated_cor_position
        if params.estimated_cor is None and scan.estimated_cor_frm_motor is not None:
            params.estimated_cor = scan.estimated_cor_frm_motor
            _logger.info(
                "{}: set estimated cor from motor position ({})".format(
                    str(scan), params.estimated_cor
                )
            )
        if scan.dim_1 is not None:
            params.image_width = scan.dim_1
        scan.saaxis_params = params
        cors_res, self._std_outs, self._std_errs = one_slice_several_cor(
            scan=scan,
            configuration=self.get_configuration(),
        )
        scan.saaxis_params.scores = cors_res
        self.autofocus(scan=scan)
        self._process_end(scan=scan, cors_res=cors_res)
        return scan

    def _process_end(self, scan, cors_res):
        assert isinstance(scan, TomwerScanBase)
        try:
            extra = {
                logconfig.DOC_TITLE: self._scheme_title,
                logconfig.SCAN_ID: str(scan),
            }
            slice_index = self.get_configuration().get("slice_index", None)

            if cors_res is None:
                info = "fail to compute cor scores of slice {} for scan {}." "".format(
                    slice_index, scan
                )
                _logger.processFailed(info, extra=extra)
                ProcessManager().notify_dataset_state(
                    dataset=scan, process=self, state=DatasetState.FAILED, details=info
                )
            else:
                info = "cor scores of slice {} for scan {} computed." "".format(
                    slice_index, scan
                )
                _logger.processSucceed(info, extra=extra)
                ProcessManager().notify_dataset_state(
                    dataset=scan,
                    process=self,
                    state=DatasetState.WAIT_USER_VALIDATION,
                    details=info,
                )
        except Exception as e:
            _logger.error(e)
        else:
            if self._dump_process:
                SAAxisProcess.process_to_tomwer_processes(
                    scan=scan,
                )

    @staticmethod
    def program_name():
        """Name of the program used for this processing"""
        return "semi-automatic axis"

    @staticmethod
    def program_version():
        """version of the program used for this processing"""
        return tomwer.version.version

    @staticmethod
    def definition():
        """definition of the process"""
        return "Semi automatic center of rotation / axis calculation"

    @staticmethod
    def process_to_tomwer_processes(scan):
        if scan.process_file is not None:
            entry = "entry"
            if isinstance(scan, HDF5TomoScan):
                entry = scan.entry

            cor = None
            if hasattr(scan, "axis_params"):
                cor = scan.axis_params.relative_cor_value

            process_index = scan.pop_process_index()
            with scan.acquire_process_file_lock():
                BaseProcess._register_process(
                    process_file=scan.process_file,
                    entry=entry,
                    results={"center_of_rotation": cor if cor is not None else "-"},
                    configuration=scan.saaxis_params.to_dict(),
                    process_index=process_index,
                    overwrite=True,
                    process=SAAxisProcess,
                )
                SAAxisProcess._extends_results(
                    scan=scan, entry=entry, process_index=process_index
                )

    @staticmethod
    def _extends_results(scan, entry, process_index):
        process_file = scan.process_file
        process_name = "tomwer_process_" + str(process_index)

        def get_process_path():
            return "/".join((entry or "entry", process_name))

        # save it to the file
        with BaseProcess._get_lock(process_file):
            # needs an extra lock for multiprocessing

            with HDF5File(process_file, mode="a") as h5f:
                nx_process = h5f.require_group(get_process_path())
                if "NX_class" not in nx_process.attrs:
                    nx_process.attrs["NX_class"] = "NXprocess"

                results = nx_process.require_group("results")
                for cor, (url, score) in scan.saaxis_params.scores.items():
                    results_cor = results.require_group(str(cor))
                    for method in ScoreMethod:
                        results_cor[method.value] = score.get(method)

                    link_path = os.path.relpath(
                        url.file_path(),
                        os.path.dirname(process_file),
                    )
                    results_cor["reconstructed_slice"] = h5py.ExternalLink(
                        link_path, url.data_path()
                    )



# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "08/02/2017"


from collections import OrderedDict

from silx.gui import qt

from tomwer.core.octaveh5editor import OctaveH5Editor
from tomwer.core.process.reconstruction.ftseries.params import fastsetupdefineglobals
import logging

logger = logging.getLogger(__name__)


class RawH5ParamEditor(OctaveH5Editor, qt.QTabWidget):
    """
    This is a simple interface allowing the user to change values of an
    h5 file

    :param QObject parent: the parent widget
    :param str h5File: the path of h5File to load on the editor
    :param float readedOctaveVersion: the version which as write the h5File
    """

    sigH5StructSettedToDefault = qt.Signal(list)
    """signal emitted when some structure are not found in the h5 file and
    set to default
    """

    def __init__(self, parent, h5File):
        OctaveH5Editor.__init__(self)
        qt.QWidget.__init__(self, parent)

        self.__initPalette = self.palette()
        self._loadH5File(h5File)
        self.defaultStructs = fastsetupdefineglobals.getAllDefaultStructures()

    def _loadH5File(self, h5File):
        """Main function to build the GUI"""
        self.h5File = h5File
        if self.h5File is None:
            return

        reader = fastsetupdefineglobals.FastSetupAll()
        reader.readAll(self.h5File, 3.8)
        self.loadReconsParams(reader.structures)

    def getInitialPalette(self):
        return self.__initPalette

    def getPalette(self, highlight=False):
        p = self.getInitialPalette()
        brush = qt.QBrush()
        p.setBrush(self.backgroundRole(), brush)

        if highlight:
            p.setColor(qt.QPalette.WindowText, qt.Qt.red)
            p.setColor(qt.QPalette.Text, qt.Qt.red)
            p.setColor(qt.QPalette.Window, qt.Qt.red)
        else:
            p.setColor(qt.QPalette.WindowText, qt.Qt.black)
            p.setColor(qt.QPalette.Text, qt.Qt.black)
            p.setColor(qt.QPalette.Window, qt.Qt.black)
        return p

    def getDefaultStruct(self, structID):
        if structID not in self.defaultStructs:
            raise ValueError("%s structure has no default value" % structID)
        else:
            return self.defaultStructs[structID]

    # implementation of OctaveH5Editor function
    def getStructs(self):
        """Return a list of all the structures with their values"""
        res = {}
        for struct in self.structs:
            fields = {}
            for field in self.structs[struct]:
                self.structs[struct][field].text()
                fields[field] = self.structs[struct][field].text()
            res[struct] = fields

        return res

    def loadReconsParams(self, structures):
        structureIDs = ["FT", "PYHSTEXE", "FTAXIS", "PAGANIN", "BEAMGEO", "DKRF"]
        self.structs = {}
        settedToDefault = []
        for structID in structureIDs:
            highlightStruct = False
            highlightedVariables = []
            try:
                struct = structures[structID]
                FSDG = fastsetupdefineglobals.FastSetupAll()
                for variable in FSDG.defaultValues[structID]:
                    if variable not in struct:
                        highlightedVariables.append(variable)
                        struct[variable] = FSDG.defaultValues[structID][variable]
                        logger.info("%s variable in %s is missing" % (variable, struct))
            except (ValueError):
                highlightStruct = True
                struct = self.getDefaultStruct(structID)
                settedToDefault.append(structID)
                warn = (
                    "%s structure not found in the given file, values setted to default"
                    % struct
                )
                logger.info(warn)

            self.structs[structID] = self._addStruct(
                OrderedDict(sorted(struct.items(), key=lambda t: t[0])),
                structID,
                highlight=highlightStruct,
                highlightedVariables=highlightedVariables,
            )

        # for now we are only displaying a message box if some structure are
        # setted to default values. Should this evolve ? If so a signal is emitted
        # and can be used back
        if len(settedToDefault) > 0:
            self.sigH5StructSettedToDefault.emit(settedToDefault)
            msg = qt.QMessageBox(self)
            msg.setIcon(qt.QMessageBox.Critical)
            text = (
                "The following structure haven't been found in %s, you might check if setted values are corrects"
                % self.h5File
            )
            for s in settedToDefault:
                text += "\n    - %s" % s
            msg.setText(text)
            msg.exec_()

    def _addStruct(self, struct, structID, highlight, highlightedVariables):
        """Add the given structure to the GUI"""
        newStruct = qt.QWidget()
        newStruct.setLayout(qt.QGridLayout())
        newStruct.setPalette(self.getPalette(highlight))
        self.addTab(newStruct, structID)
        dicoFieldToQLE = {}
        for i, field in enumerate(struct):
            dicoFieldToQLE[field] = self._addField(
                field, struct[field], newStruct, i, field in highlightedVariables
            )

        return OrderedDict(sorted(dicoFieldToQLE.items(), key=lambda t: t[0]))

    def _addField(self, field, value, parent, index, highlighted=False):
        """Add the field to the widget

        :param str field: the ID of the new field to defined
        :param value: the value of the field
        :param parent: the Qt parent
        :param index: the index of the field
        :return: the line edit for the field
        """
        titleItem = qt.QLabel(field)
        titleItem.setPalette(self.getPalette(highlighted))
        parent.layout().addWidget(titleItem, index, 0)
        lineEdit = qt.QLineEdit(parent=parent, text=str(value))
        parent.layout().addWidget(lineEdit, index, 1)
        return lineEdit



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "04/03/2019"


from collections import OrderedDict
import datetime
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.blissscan import BlissScan


def _assert_param_instance(param, valid_instances):
    """check the parameter value, can be an instance of valid_instance or
    an iterable of valid_instance"""
    if isinstance(param, valid_instances):
        return
    else:
        try:
            for unique_param in param:
                if not isinstance(unique_param, valid_instances):
                    assert isinstance(unique_param, valid_instances)
        except:
            raise ValueError("%s has an invalid type" % param)


def _assert_cast_to_boolean(value):
    """

    :param value: value we want to cast to a boolean
    """
    try:
        bool(int(value))
    except:
        raise ValueError("%s cannot be cast to boolean" % value)


def _assert_cast_to_int(value):
    try:
        int(value)
    except:
        raise ValueError("%s cannot be cast to int" % value)


class LastReceivedScansDict(OrderedDict):
    """List the received scan from the first received to the last received"""

    def __init__(self, limit=None):
        """Simple structure in order to store the received last elements and
        the time of acquisition
        """
        assert limit is None or (type(limit) is int and limit > 0)
        OrderedDict.__init__(self)
        self.limit = limit

    def add(self, scan):
        assert isinstance(scan, (TomwerScanBase, BlissScan))
        self[str(scan)] = datetime.datetime.now()
        if self.limit is not None and len(self) > self.limit:
            self.pop(list(self.keys())[0])


class IgnoreProcess:
    """Simple util class to ignore a processing when using pypushflow"""

    def process(self, scan):
        return scan

    def set_properties(self, properties):
        pass

    __call__ = process



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "09/06/2017"

import os
import shutil
import tempfile
import time
import unittest

from silx.gui.utils.testutils import TestCaseQt
from silx.gui import qt

from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.gui.control.datawatcher import DataWatcherWidget
from tomwer.gui.control.datawatcher.datawatcherobserver import _QDataWatcherFixObserver
from tomwer.gui.control.datawatcher.datawatcherobserver import _QDataWatcherObserver
from tomwer.gui.control.datawatcher.datawatcherobserver import _QOngoingObservation
from tomwer.core.process.control.datawatcher import status as datawatcherstatus
from tomwer.synctools.rsyncmanager import RSyncManager
from tomwer.test.utils import UtilsTest
from tomwer.test.utils import skip_gui_test


@unittest.skip("Fail on CI...")
@unittest.skipIf(skip_gui_test(), reason="skip gui test")
class TestDataWatcherDirObserver(TestCaseQt):
    """
    Simple test to make sure the timeout of data watcher is working properly
    """

    FOLDER_WITH_DATA = 4

    FOLDER_WITHOUT_DATA = 3

    def setUp(self):
        TestCaseQt.setUp(self)
        self.observeFolder = tempfile.mkdtemp()

        self.dataSetID = "test10"
        dataDir = UtilsTest.getEDFDataset(self.dataSetID)
        for iFolder in range(self.FOLDER_WITH_DATA):
            shutil.copytree(
                dataDir,
                os.path.join(self.observeFolder, "f" + str(iFolder), self.dataSetID),
            )

        for iFolder in range(self.FOLDER_WITHOUT_DATA):
            os.makedirs(os.path.join(self.observeFolder, "empty" + str(iFolder)))

        # observer
        self.observer = _QDataWatcherObserver(
            headDir=self.observeFolder,
            startByOldest=False,
            obsMethod=datawatcherstatus.DET_END_XML,
            srcPattern=None,
            destPattern=None,
            observationClass=_QOngoingObservation,
        )

    def tearDown(self):
        self.observer.wait()
        self.observer = None
        if os.path.isdir(self.observeFolder):
            shutil.rmtree(self.observeFolder)
        TestCaseQt.tearDown(self)

    def testRun(self):
        self.observer.run()
        self.assertTrue(len(self.observer.observations) is self.FOLDER_WITH_DATA)
        self.observer.waitForObservationFinished(4)


@unittest.skipIf(skip_gui_test(), reason="skip gui test")
class TestDataWatcherFixObserver(TestCaseQt):
    def setUp(self):
        TestCaseQt.setUp(self)

        self.observeFolder = tempfile.mkdtemp()
        self.dataSetID = "test10"
        dataDir = UtilsTest.getEDFDataset(self.dataSetID)
        shutil.copytree(dataDir, os.path.join(self.observeFolder, self.dataSetID))

        self.thread = _QDataWatcherFixObserver(
            scanID=os.path.join(self.observeFolder, self.dataSetID),
            obsMethod=datawatcherstatus.DET_END_XML,
            srcPattern=None,
            destPattern=None,
            patternObs=None,
            observationRegistry=None,
        )

    def tearDown(self):
        self.thread.wait()
        self.thread = None
        shutil.rmtree(self.observeFolder)
        TestCaseQt.tearDown(self)

    def testRunOk(self):
        self.thread.start()
        self.thread.wait(5)

    def testFolderNotExisting(self):
        self.thread.setDirToObserve("toto/pluto")
        self.thread.start()
        self.thread.wait(5)
        self.assertEqual(
            self.thread.status, "failure", msg="status is %s" % self.thread.status
        )

    def testRunWaitingXML(self):
        fileXML = os.path.join(
            self.thread.dataWatcherProcess.RootDir, self.dataSetID + ".xml"
        )
        assert os.path.isfile(fileXML)
        os.remove(fileXML)
        self.thread.start()
        self.thread.wait(5)


class _ObservationCounter(qt.QObject):
    """Basically simulate the next box which has to receive some signal to
    process"""

    def __init__(self):
        qt.QObject.__init__(self)
        self.scansCounted = 0
        self.scansID = []

    def add(self, scanID):
        self.scansCounted = self.scansCounted + 1
        if scanID.path not in self.scansID:
            self.scansID.append(scanID.path)

    def getReceivedScan(self):
        return self.scansCounted

    def getDifferentScanReceived(self):
        return len(self.scansID)

    def clear(self):
        self.scansID = []
        self.scansCounted = 0


@unittest.skip("Fail on CI...")
@unittest.skipIf(skip_gui_test(), reason="skip gui test")
class TestDataWatcher(TestCaseQt):
    """
    Make sur the DataWatcher process is valid
    """

    WAIT_BTW_LOOP = 0.1

    def setUp(self):
        TestCaseQt.setUp(self)
        # create a folder with an unfinished acquisition
        self.observeFolder = tempfile.mkdtemp()

        self.dataSetIDs = "test10", "test01"
        for scanID in self.dataSetIDs:
            dataDir = UtilsTest.getEDFDataset(scanID)
            shutil.copytree(dataDir, os.path.join(self.observeFolder, scanID))

        self.observationCounter = _ObservationCounter()

        # data watcher
        self.dataWatcher = DataWatcherWidget()
        self.dataWatcher.setFolderObserved(self.observeFolder)
        self.dataWatcher.obsMethod = datawatcherstatus.DET_END_XML
        self.dataWatcher.setSrcAndDestPattern(srcPattern=None, destPattern=None)
        self.dataWatcher.setWaitTimeBtwLoop(self.WAIT_BTW_LOOP)
        self.dataWatcher.sigScanReady.connect(self.observationCounter.add)

    def tearDown(self):
        self.dataWatcher.stop()
        self.dataWatcher = None
        if os.path.isdir(self.observeFolder):
            shutil.rmtree(self.observeFolder)
        TestCaseQt.tearDown(self)

    def testInfinity(self):
        """
        Test that if we ar ein the infinity mode and two acquisitions are
        existing we will found only those two observations once and that the
        datawatcher object is still observing
        """
        self.observationCounter.clear()

        self.dataWatcher.start()
        self.qapp.processEvents()
        time.sleep(self.WAIT_BTW_LOOP * 2.0)  # make sure no multiple signal are emitted
        self.qapp.processEvents()
        time.sleep(0.4)
        self.qapp.processEvents()
        self.dataWatcher.waitForObservationFinished()
        self.assertTrue(self.dataWatcher.observationThread is not None)
        self.assertTrue(len(self.dataWatcher.lastFoundScans) == 2)
        self.assertTrue(self.observationCounter.getReceivedScan() == 2)
        self.assertTrue(self.observationCounter.getDifferentScanReceived() == 2)
        self.dataWatcher.stop()

    def testPatternFileObsMethod(self):
        """
        Test the DET_END_USER_ENTRY observation method.
        Will look for a file pattern given by the user in the scan directory
        """

        def tryPattern(pattern, filePrefix, fileSuffix):
            self.dataWatcher.stop()
            self.observationCounter.clear()

            self.dataWatcher.setObsMethod(
                (datawatcherstatus.DET_END_USER_ENTRY, {"pattern": pattern})
            )

            tempfile.mkdtemp(
                prefix=filePrefix,
                suffix=fileSuffix,
                dir=os.path.join(self.observeFolder, "test10"),
            )

            self.assertEqual(self.observationCounter.getDifferentScanReceived(), 0)

            self.dataWatcher.start()
            self.dataWatcher.waitForObservationFinished()
            self.qapp.processEvents()
            time.sleep(0.4)
            self.qapp.processEvents()

            return self.observationCounter.getDifferentScanReceived() is 1

        self.assertTrue(
            tryPattern(
                pattern="*tatayoyo*.cfg", filePrefix="tatayoyo", fileSuffix=".cfg"
            )
        )

        self.assertFalse(
            tryPattern(pattern="*tatayoyo*.cfg", filePrefix="tatayoyo", fileSuffix="")
        )

        self.assertFalse(
            tryPattern(pattern="tatayoyo.cfg", filePrefix="tatayoyo", fileSuffix=".cfg")
        )


@unittest.skipIf(skip_gui_test(), reason="skip gui test")
class TestDataWatcherAborted(TestCaseQt):
    """Test behavior of the DataWatcher when some acquisition are aborted"""

    def setUp(self):
        TestCaseQt.setUp(self)
        # create a folder with an unfinished acquisition
        self.observeFolder = tempfile.mkdtemp()

        # data watcher
        self.dataWatcher = DataWatcherWidget()
        self.dataWatcher.setFolderObserved(self.observeFolder)
        self.dataWatcher.obsMethod = datawatcherstatus.DET_END_XML
        self.dataWatcher.setSrcAndDestPattern(srcPattern=None, destPattern=None)

        self.dataDir01 = UtilsTest.getEDFDataset("test01")
        self.dataDir10 = UtilsTest.getEDFDataset("test10")

        RSyncManager().force_sync = False

    def tearDown(self):
        self.dataWatcher.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.dataWatcher.stop()
        self.dataWatcher = None
        if os.path.exists(self.observeFolder):
            shutil.rmtree(self.observeFolder)
        TestCaseQt.tearDown(self)

    def testAbortedFindNext(self):
        """Make sure if an acquisition is aborted then we will skip it
        and that the NEXT acquisition will be found"""
        self.dataWatcher.setWaitTimeBtwLoop(0.1)
        self.assertEqual(len(self.dataWatcher.getIgnoredFolders()), 0)

        dirTest01 = os.path.join(self.observeFolder, "test01")
        shutil.copytree(self.dataDir01, dirTest01)
        # add the `abort file`
        open(os.path.join(dirTest01, "test01" + EDFTomoScan.ABORT_FILE), "a")
        self.dataWatcher.start()
        self.dataWatcher.observationThread.wait()
        self.qapp.processEvents()
        self.assertEqual(len(self.dataWatcher.getIgnoredFolders()), 0)
        self.assertTrue(self.dataWatcher.isObserving is True)
        dirTest10 = os.path.join(self.observeFolder, "test10")
        shutil.copytree(self.dataDir10, dirTest10)

        for i in range(6):
            self.qapp.processEvents()
            time.sleep(0.1)
        self.assertEqual(len(self.dataWatcher.lastFoundScans), 1)
        self.assertTrue(dirTest10 in self.dataWatcher.lastFoundScans)

    def testAbortedFindPrevious(self):
        """Make sure if an acquisition is aborted then we will skip it
        and that the PREVIOUS acquisition will be found"""
        self.dataWatcher.setWaitTimeBtwLoop(0.1)
        self.assertEqual(len(self.dataWatcher.getIgnoredFolders()), 0)

        dirTest10 = os.path.join(self.observeFolder, "test10")
        shutil.copytree(self.dataDir10, dirTest10)
        dirTest01 = os.path.join(self.observeFolder, "test01")
        shutil.copytree(self.dataDir01, dirTest01)
        # add the `abort file`
        open(os.path.join(dirTest01, "test01" + EDFTomoScan.ABORT_FILE), "a")
        self.dataWatcher.start()
        self.dataWatcher.observationThread.wait()
        self.qapp.processEvents()
        self.assertTrue(self.dataWatcher.isObserving is True)
        time.sleep(0.2)
        self.qapp.processEvents()
        time.sleep(0.2)
        self.qapp.processEvents()
        self.assertEqual(len(self.dataWatcher.lastFoundScans), 1)
        self.assertTrue(dirTest10 in self.dataWatcher.lastFoundScans)

    def testAbortedWhenWaitingFor(self):
        """
        Make sure will skip the scan if aborted even if observation started
        """
        self.dataWatcher.setWaitTimeBtwLoop(0.1)
        self.assertEqual(len(self.dataWatcher.getIgnoredFolders()), 0)

        dirTest01 = os.path.join(self.observeFolder, "test01")
        shutil.copytree(self.dataDir01, dirTest01)
        self.dataWatcher.start()
        # add the `abort file`
        open(os.path.join(dirTest01, "test01" + EDFTomoScan.ABORT_FILE), "a")
        self.dataWatcher.observationThread.wait()
        self.qapp.processEvents()
        self.assertEqual(len(self.dataWatcher.getIgnoredFolders()), 0)
        self.assertTrue(self.dataWatcher.isObserving)
        dirTest10 = os.path.join(self.observeFolder, "test10")
        shutil.copytree(self.dataDir10, dirTest10)
        self.qapp.processEvents()
        time.sleep(0.2)
        self.qapp.processEvents()
        self.assertTrue(len(self.dataWatcher.lastFoundScans) is 1)
        self.assertTrue(dirTest10 in self.dataWatcher.lastFoundScans)

    def testAbortedRemovingFolder(self):
        """Make sure the aborted folder is removed after found by datawatcher"""
        RSyncManager().force_sync = True
        self.dataWatcher.setWaitTimeBtwLoop(0.1)
        dirTest01 = os.path.join(self.observeFolder, "test01")
        shutil.copytree(self.dataDir01, dirTest01)
        # add the `abort file`
        open(os.path.join(dirTest01, "test01" + EDFTomoScan.ABORT_FILE), "a")
        self.dataWatcher.start()
        self.dataWatcher.observationThread.wait()
        assert dirTest01 in self.dataWatcher.observationThread.observations.dict
        obs = self.dataWatcher.observationThread.observations.dict[dirTest01]
        obs.wait()
        self.assertFalse(os.path.exists(dirTest01))


def suite():
    test_suite = unittest.TestSuite()
    for ui in (
        TestDataWatcherDirObserver,
        TestDataWatcher,
        TestDataWatcherFixObserver,
        TestDataWatcherAborted,
    ):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "11/02/2020"


import os
from silx.gui import qt
from tomwer.gui.reconstruction.nabu.nabuconfig.base import _NabuStageConfigBase
from tomwer.core.process.reconstruction.nabu.utils import _NabuStages
from silx.utils.enum import Enum as _Enum
import logging
from tomwer.io.utils import get_default_directory

try:
    import glymur
except ImportError:
    has_glymur = False
else:
    has_glymur = True

_logger = logging.getLogger(__name__)


class _NabuOutputFileFormat(_Enum):
    # NPY = 'npy'
    # NPZ = 'npz'
    TIFF = "tiff"
    HDF5 = "hdf5"
    JP2K = "jp2k"
    EDF = "edf"


class _NabuOutputConfig(_NabuStageConfigBase, qt.QWidget):
    """
    Widget to define the output configuration of nabu
    """

    sigConfChanged = qt.Signal(str)
    """Signal emitted when the configuration change. Parameter is the option
    modified
    """

    def __init__(self, parent):
        _NabuStageConfigBase.__init__(self, stage=_NabuStages.POST)
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())

        # output dir
        self._output_dir_widget = qt.QWidget(self)
        self._outputLabel = qt.QLabel("output:", self._output_dir_widget)
        self.layout().addWidget(self._output_dir_widget, 0, 0, 1, 4)
        self._output_dir_widget.setLayout(qt.QHBoxLayout())
        self._output_dir_widget.layout().setContentsMargins(0, 0, 0, 0)
        self._output_dir_widget.layout().setSpacing(0)
        self._output_dir_widget.layout().addWidget(self._outputLabel)
        self._defaultOutput = qt.QCheckBox(
            "raw data root level", self._output_dir_widget
        )
        self._defaultOutput.setToolTip(
            "set reconstruction at the the same " "level as raw data"
        )
        self._output_dir_widget.layout().addWidget(self._defaultOutput)
        self._outputDirQLE = qt.QLineEdit("", self._output_dir_widget)
        self._output_dir_widget.layout().addWidget(self._outputDirQLE)
        style = qt.QApplication.style()
        icon_opendir = style.standardIcon(qt.QStyle.SP_DirOpenIcon)
        self._selectOutputPB = qt.QPushButton(icon_opendir, "", self._output_dir_widget)
        self._selectOutputPB.setIcon(icon_opendir)
        self._selectOutputPB.setToolTip("select output directory")
        self._output_dir_widget.layout().addWidget(self._selectOutputPB)
        # is required in nabu configuration file but tomwer wan't to force
        # it to be in expert mode
        self.registerWidget(self._output_dir_widget, "advanced")

        # file format
        self._outputFileFormatLabel = qt.QLabel("output file format:", self)
        self.layout().addWidget(self._outputFileFormatLabel, 1, 0, 1, 1)
        self._fileFormatCB = qt.QComboBox(self)
        for ff in _NabuOutputFileFormat:
            if ff is _NabuOutputFileFormat.JP2K:
                if not has_glymur:
                    _logger.warning(
                        "could not load jp2k format, glymur and OpenJPEG requested"
                    )
                else:
                    from glymur import version

                    if version.openjpeg_version < "2.3.0":
                        _logger.warning(
                            "You must have at least version 2.3.0 of OpenJPEG "
                            "in order to write jp2k images."
                        )
                    else:
                        self._fileFormatCB.addItem(ff.value)
            else:
                self._fileFormatCB.addItem(ff.value)
        self.layout().addWidget(self._fileFormatCB, 1, 2, 1, 2)
        self.registerWidget(self._outputFileFormatLabel, "optional")
        self.registerWidget(self._fileFormatCB, "optional")

        # file per group
        self._filePerGroupLabel = qt.QLabel("frame per group:", self)
        self.layout().addWidget(self._filePerGroupLabel, 2, 0, 1, 1)
        self._framePerGroup = qt.QSpinBox(self)
        self._framePerGroup.setMinimum(100)
        self._framePerGroup.setSingleStep(50)
        self._framePerGroup.setMaximum(10000)
        # not managed for now so hide
        self._filePerGroupLabel.hide()
        self._framePerGroup.hide()
        self.layout().addWidget(self._framePerGroup, 2, 2, 1, 2)

        # spacer for style
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer, 200, 0, 1, 1)

        # set up
        self._defaultOutput.setChecked(True)
        self._hideOutput(True)
        self.setFramePerGroup(100)
        self.setFileformat("hdf5")

        # connect signal / slot
        self._defaultOutput.toggled.connect(self._hideOutput)
        self._selectOutputPB.released.connect(self._selectOutput)
        self._outputDirQLE.editingFinished.connect(self._outputDirChanged)
        self._defaultOutput.toggled.connect(self._outputDirChanged)
        self._fileFormatCB.currentTextChanged.connect(self._fileFormatChanged)
        self._framePerGroup.valueChanged.connect(self._framePerGroupChanged)
        self._hideOutput(True)

    def _hideOutput(self, hide):
        self._outputDirQLE.setVisible(not hide)
        self._selectOutputPB.setVisible(not hide)

    def _selectOutput(self):
        defaultDirectory = self._outputDirQLE.text()
        if os.path.isdir(defaultDirectory):
            defaultDirectory = get_default_directory()

        dialog = qt.QFileDialog(self, directory=defaultDirectory)
        dialog.setFileMode(qt.QFileDialog.DirectoryOnly)

        if not dialog.exec_():
            dialog.close()
            return

        self._outputDirQLE.setText(dialog.selectedFiles()[0])
        self._outputDirChanged()

    def _outputDirChanged(self):
        self.sigConfChanged.emit("location")

    def _fileFormatChanged(self):
        self.sigConfChanged.emit("file_format")

    def _framePerGroupChanged(self):
        self.sigConfChanged.emit("frames_per_group")

    def getOutputDir(self):
        """

        :return: None if the default output directory is selected else
                 return path to the directory
        """
        if self._defaultOutput.isChecked():
            return None
        else:
            return self._outputDirQLE.text()

    def setOutputDir(self, dir):
        if dir in (None, ""):
            self._defaultOutput.setChecked(True)
        else:
            self._defaultOutput.setChecked(False)
            self._outputDirQLE.setText(dir)

    def getFileFormat(self) -> _NabuOutputFileFormat:
        return _NabuOutputFileFormat.from_value(self._fileFormatCB.currentText())

    def setFileformat(self, file_format):
        file_format = _NabuOutputFileFormat.from_value(file_format)
        index = self._fileFormatCB.findText(file_format.value)
        self._fileFormatCB.setCurrentIndex(index)

    def getFramePerGroup(self):
        return self._framePerGroup.value()

    def setFramePerGroup(self, n_frames):
        self._framePerGroup.setValue(n_frames)

    def getConfiguration(self):
        return {
            "file_format": self.getFileFormat().value,
            "location": self.getOutputDir() or "",
            # 'frames_per_group': self.getFramePerGroup(),
        }

    def setConfiguration(self, config):
        if "file_format" in config:
            self.setFileformat(config["file_format"])
        if "location" in config:
            self.setOutputDir(config["location"])
        if "frames_per_group" in config:
            self.setFramePerGroup(int(config["frames_per_group"]))


if __name__ == "__main__":
    app = qt.QApplication([])
    widget = _NabuOutputConfig(parent=None)
    widget.show()
    print(widget.getConfiguration())
    app.exec_()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "09/02/2018"

import logging
import os
import shutil
import tempfile
from glob import glob
import unittest

from orangecontrib.tomwer.test.utils import OrangeWorflowTest
from tomwer.gui.qtapplicationmanager import QApplicationManager
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.test.utils import UtilsTest
from tomwer.test.utils import skip_gui_test, skip_orange_workflows_test

logging.disable(logging.INFO)

app = QApplicationManager()


@unittest.skipIf(skip_orange_workflows_test(), reason="skip orange workflows tests")
@unittest.skipIf(skip_gui_test(), reason="skip gui test")
class TestNabuWidget(OrangeWorflowTest):
    """
    test the workflow composed of the following widgets :
        - DataList
        - nabu reconstruction
        - data validator
    """

    @classmethod
    def setUpClass(cls):
        OrangeWorflowTest.setUpClass()
        # create widgets
        cls.nodeDataList = cls.addWidget(
            cls, "orangecontrib.tomwer.widgets.control.DataListOW.DataListOW"
        )
        cls.nodeNabu = cls.addWidget(
            cls, "orangecontrib.tomwer.widgets.reconstruction.NabuOW.NabuOW"
        )
        cls.nodeValidator = cls.addWidget(
            cls, "orangecontrib.tomwer.widgets.control.DataValidatorOW.DataValidatorOW"
        )

        # Let Orange process events (node creations)
        cls.processOrangeEvents(cls)

        # linking the workflow
        cls.link(cls, cls.nodeDataList, "data", cls.nodeNabu, "data")
        cls.link(cls, cls.nodeNabu, "data", cls.nodeValidator, "data")
        cls.link(
            cls,
            cls.nodeValidator,
            "change recons params",
            cls.nodeNabu,
            "change recons params",
        )

        # Let Orange process events (node creations)
        cls.processOrangeEvents(cls)

        # getting the widgets
        cls.dataListWidget = cls.getWidgetForNode(cls, cls.nodeDataList)
        cls.nabuWidget = cls.getWidgetForNode(cls, cls.nodeNabu)
        cls.dataValidatorWidget = cls.getWidgetForNode(cls, cls.nodeValidator)

    @classmethod
    def tearDownClass(cls):
        for node in (cls.nodeDataList, cls.nodeNabu, cls.nodeValidator):
            cls.removeNode(cls, node)
        del cls.dataListWidget
        del cls.nabuWidget
        del cls.dataValidatorWidget
        OrangeWorflowTest.tearDownClass()

    def setUp(self):
        OrangeWorflowTest.setUp(self)
        dataset = "D2_H2_T2_h_"
        self.folder = tempfile.mkdtemp()
        dataset_folder = os.path.join(self.folder, dataset)
        shutil.copytree(src=UtilsTest().getEDFDataset(dataset), dst=dataset_folder)
        self.dataListWidget.add(dataset_folder)
        self.scan = ScanFactory.create_scan_object(scan_path=dataset_folder)
        # make sure there is no .cfg file inside
        assert len(self.getCfgFiles()) == 0

        self.nabuWidget.setDryRun(True)
        self.nabuWidget.setConfiguration({"tomwer_slices": None})

        # monkey patch exec_ to check behavior
        self.nabuWidget._replaceExec_()

    def tearDown(self):
        shutil.rmtree(self.folder)
        OrangeWorflowTest.tearDown(self)

    def testWorkflow(self):
        self.dataListWidget._sendList()

        # wait until the reconstruction is made and the data watcher pop
        self.qWaitForWindowExposed(self.dataValidatorWidget, timeout=10)

        # check .cfg files are here and validator is treating the current scan
        for i in range(10):
            self.qapp.processEvents()
            import time

            time.sleep(0.1)
            self.qapp.processEvents()
        current_scan_path = self.dataValidatorWidget.getCurrentScanID()
        self.assertEqual(current_scan_path, self.scan.path)
        self.assertEqual(len(self.getCfgFiles()), 1)
        # validate the scan
        self.dataValidatorWidget.changeReconsParamCurrentScan()
        self.qapp.processEvents()

        # wait for ftseries to be pop up within the dialog
        self.qWaitForWindowExposed(self.nabuWidget, timeout=10)

    def getCfgFiles(self):
        return glob(self.scan.path + os.sep + "*.cfg")


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestNabuWidget,):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "01/08/2018"

from Orange.widgets import widget, gui
from Orange.widgets.widget import Input
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.visualization.sinogramviewer import SinogramViewer
from ..utils import WidgetLongProcessing
import logging

logger = logging.getLogger(__name__)


class SinogramViewerOW(WidgetLongProcessing, widget.OWWidget):
    """
    This widget can be used to compute and display a specific sinogram from an
    acquisition.
    """

    name = "sinogram viewer"
    id = "orange.widgets.tomwer.visualization.sinogramviewer"
    description = (
        "This widget can be used to compute and display a "
        "specific sinogram from an acquisition."
    )

    icon = "icons/sinogramviewer.png"
    priority = 5
    category = "tomwer"
    keywords = ["tomography", "sinogram", "radio"]

    allows_cycle = True
    compress_signal = False

    want_main_area = True
    resizing_enabled = True

    class Inputs:
        data_in = Input(name="data", type=TomwerScanBase)

    def __init__(self, parent=None):
        widget.OWWidget.__init__(self, parent)
        WidgetLongProcessing.__init__(self)
        self._box = gui.vBox(self.mainArea, self.name)
        self._viewer = SinogramViewer(parent=self)
        self._box.layout().addWidget(self._viewer)

        # connect signal / slot
        self._viewer.sigSinoLoadStarted.connect(self._startProcessing)
        self._viewer.sigSinoLoadEnded.connect(self._endProcessing)
        self.destroyed.connect(self._viewer._stopAnimationThread)

    @Inputs.data_in
    def addLeafScan(self, scanID):
        if scanID is None:
            return
        self._viewer.setScan(scanID)



"""Utilities for with-statement contexts.  See PEP 343."""
import abc
import _collections_abc

__all__ = ["AbstractContextManager"]


class AbstractContextManager(abc.ABC):

    """An abstract base class for context managers."""

    def __enter__(self):
        """Return `self` upon entering the runtime context."""
        return self

    @abc.abstractmethod
    def __exit__(self, exc_type, exc_value, traceback):
        """Raise any exception triggered within the runtime context."""
        return None

    @classmethod
    def __subclasshook__(cls, C):
        if cls is AbstractContextManager:
            return _collections_abc._check_methods(C, "__enter__", "__exit__")
        return NotImplemented



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

"""
This module is used to define the process of the reference creator.
This is related to the issue #184
"""

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "19/07/2018"

import fnmatch
import logging
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
import os
import re

from tomwer.core.process.baseprocess import SingleProcess, _input_desc, _output_desc

_logger = logging.getLogger(__name__)


class BaseFilter(object):
    """
    Apply a filter to an object
    """

    def description(self):
        pass

    def isFiltered(self, value):
        """
        Return True if the value not filtered

        """
        raise NotImplementedError("Base class")


class _PatternBaseFilter(BaseFilter):
    """Filter based on a pattern"""

    def __init__(self, pattern):
        BaseFilter.__init__(self)
        self.setPattern(pattern)

    def setPattern(self, pattern):
        """
        compile th filter for the given pattern
        :param str pattern:
        """
        self._pattern = pattern

    def getPattern(self):
        return self._pattern


class RegularExpressionFilter(_PatternBaseFilter):
    """Filter a string based on a defined pattern"""

    def __init__(self, pattern):
        _PatternBaseFilter.__init__(self, pattern=pattern)

    def setPattern(self, pattern):
        """
        compile th filter for the given pattern
        :param str pattern:
        """
        super().setPattern(pattern)
        if self._pattern is not None:
            try:
                self._filter = re.compile(self._pattern)
            except re.error as e:
                self.unvalidPatternDefinition(self._pattern, e)
                _logger.error(e)

    def description(self):
        return "Filter a string base on a regular expression"

    def isFiltered(self, value):
        try:
            match = self._filter.match(value) is None
        except:
            return False
        else:
            return match

    def unvalidPatternDefinition(self, pattern, error):
        _logger.error("%s is not a valid pattern. Error is %s" % (pattern, error))


class UnixFileNamePatternFilter(_PatternBaseFilter):
    """Filter a string based on 'fnmatch' module (unix filename pattern
    matching)"""

    def __init__(self, pattern):
        _PatternBaseFilter.__init__(self, pattern)

    def description(self):
        return "Filter a string base on a glob (unix style pathname)"

    def isFiltered(self, value):
        try:
            match = fnmatch.fnmatch(value, self._pattern)
        except:
            match = False
        return not match


class FileNameFilter(_PatternBaseFilter, SingleProcess):
    """Filter than can call several filter type"""

    inputs = [
        _input_desc(
            name="data", type=TomwerScanBase, handler="process", doc="scan path"
        )
    ]

    outputs = [_output_desc(name="data", type=TomwerScanBase, doc="scan path")]

    FILTER_TYPES = "unix file name pattern", "regular expression"

    _DEFAULT_FILTER_TYPE = FILTER_TYPES[0]

    def __init__(self, pattern):
        self._fnFilter = UnixFileNamePatternFilter(pattern)
        self._reFilter = RegularExpressionFilter(pattern)
        self._filters = {}
        self._filters["unix file name pattern"] = self._fnFilter
        self._filters["regular expression"] = self._reFilter
        self._activeFilter = self._DEFAULT_FILTER_TYPE

        _PatternBaseFilter.__init__(self, pattern)
        SingleProcess.__init__(self)
        self.activeFilter = self._DEFAULT_FILTER_TYPE

    @property
    def activeFilter(self):
        return self._activeFilter

    @activeFilter.setter
    def activeFilter(self, filter_type):
        assert filter_type in self.FILTER_TYPES
        self._activeFilter = filter_type
        self._filters[self.activeFilter].setPattern(pattern=self.getPattern())

    def setPattern(self, pattern):
        super().setPattern(pattern)
        self._filters[self._activeFilter].setPattern(pattern)

    def isFiltered(self, value):
        return self._filters[self._activeFilter].isFiltered(value)

    def process(self, scan):
        _scan = scan
        if isinstance(scan, dict):
            _scan = ScanFactory.create_scan_object_frm_dict(scan)
        if _scan is None:
            return False
        assert isinstance(_scan, TomwerScanBase)

        if not self.isFiltered(os.path.basename(_scan.path)):
            if self._return_dict:
                return _scan.to_dict()
            else:
                return _scan
        else:
            return



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import argparse
from silx.gui import qt
import signal
from typing import Union
from tomwer.gui import icons
from tomwer.gui.utils.splashscreen import getMainSplashScreen
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.gui.reconstruction.saaxis.saaxis import SAAxisWindow as _SAAxisWindow
from tomwer.core.process.reconstruction.axis.axis import AxisProcess, NoAxisUrl
from tomwer.core.process.reconstruction.saaxis.saaxis import SAAxisProcess
from tomwer.synctools.axis import QAxisRP
from tomwer.synctools.saaxis import QSAAxisParams
import logging
import time


logging.basicConfig(level=logging.INFO)
_logger = logging.getLogger(__name__)


class SAAxisThread(qt.QThread):
    """
    Thread to call nabu and reconstruct one slice with several cor value
    """

    def init(self, scan, configuration):
        self.scan = scan
        self._configuration = configuration

    def run(self) -> None:
        process = SAAxisProcess(process_id=None)
        process.set_properties(self._configuration)
        t0 = time.time()
        process.process(self.scan)
        print("execution time is {}".format(time.time() - t0))


class SAAxisWindow(_SAAxisWindow):
    def __init__(self, parent=None):
        self._scan = None
        super().__init__(parent)
        # thread for computing cors
        self._processingThread = SAAxisThread()
        self._processingThread.finished.connect(self._threadedProcessEnded)
        self.sigStartSinogramLoad.connect(self._callbackStartLoadSinogram)
        self.sigEndSinogramLoad.connect(self._callbackEndLoadSinogram)
        # processing for the cor estimation
        self._cor_estimation_process = AxisProcess(self.getQAxisRP())

        # hide the validate button
        self._saaxisControl._applyBut.hide()
        self.hideAutoFocusButton()

    def _launchReconstructions(self):
        if self._processingThread.isRunning():
            _logger.error(
                "a calculation is already launch. You must wait for "
                "it to end prior to launch a new one"
            )
        else:
            qt.QApplication.setOverrideCursor(qt.Qt.WaitCursor)
            self._processingThread.init(
                configuration=self.getConfiguration(), scan=self.getScan()
            )
            self._processingThread.start()

    def _threadedProcessEnded(self):
        saaxis_params = self._processingThread.scan.saaxis_params
        if saaxis_params is None:
            scores = None
        else:
            scores = saaxis_params.scores
        scan = self.getScan()
        assert scan is not None, "scan should have been set"
        self.setCorScores(
            scores, img_width=scan.dim_1, score_method=self.getScoreMethod()
        )
        if scan.saaxis_params.autofocus is not None:
            self.setCurrentCorValue(scan.saaxis_params.autofocus)
        self.showResults()
        qt.QApplication.restoreOverrideCursor()

    def _callbackStartLoadSinogram(self):
        print(
            "start loading sinogram for {}. Can take some time".format(self.getScan())
        )

    def _callbackEndLoadSinogram(self):
        print("sinogram loaded for {} loaded.".format(self.getScan()))

    def close(self) -> None:
        self._stopProcessingThread()
        super().close()

    def _stopProcessingThread(self):
        if self._processingThread:
            self._processingThread.terminate()
            self._processingThread.wait(500)
            self._processingThread = None

    def stop(self):
        self._stopProcessingThread()
        super().stop()

    def _computeEstimatedCor(self) -> Union[float, None]:
        scan = self.getScan()
        if scan is None:
            return
        _logger.info("{} - start cor estimation for".format(scan))
        qt.QApplication.setOverrideCursor(qt.Qt.WaitCursor)
        try:
            self._cor_estimation_process.compute(scan=scan, wait=True)
        except NoAxisUrl:
            qt.QApplication.restoreOverrideCursor()
            msg = qt.QMessageBox(self)
            msg.setIcon(qt.QMessageBox.Warning)
            text = (
                "Unable to find url to compute the axis, please select them "
                "from the `axis input` tab"
            )
            msg.setText(text)
            msg.exec_()
            return None
        else:
            self.setEstimatedCorPosition(
                value=scan.axis_params.relative_cor_value,
            )
            qt.QApplication.restoreOverrideCursor()
            self.getAutomaticCorWindow().hide()
            return scan.axis_params.relative_cor_value


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "scan_path",
        help="For EDF acquisition: provide folder path, for HDF5 / nexus"
        "provide the master file",
        default=None,
    )
    parser.add_argument(
        "--entry", help="For Nexus files: entry in the master file", default=None
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Set logging system in debug mode",
    )

    options = parser.parse_args(argv[1:])

    if options.debug:
        logging.root.setLevel(logging.DEBUG)

    global app  # QApplication must be global to avoid seg fault on quit
    app = qt.QApplication.instance() or qt.QApplication([])
    splash = getMainSplashScreen()
    qt.QApplication.setOverrideCursor(qt.Qt.WaitCursor)
    qt.QApplication.processEvents()

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler

    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catch
    timer.timeout.connect(lambda: None)

    if options.scan_path is not None:
        if os.path.isdir(options.scan_path):
            options.scan_path = options.scan_path.rstrip(os.path.sep)
            scan = ScanFactory.create_scan_object(scan_path=options.scan_path)
        else:
            if not os.path.isfile(options.scan_path):
                raise ValueError(
                    "scan path should be a folder containing an"
                    " EDF acquisition or an hdf5 - nexus "
                    "compliant file"
                )
            if options.entry is None:
                raise ValueError("entry in the master file should be specify")
            scan = HDF5TomoScan(scan=options.scan_path, entry=options.entry)
    else:
        scan = ScanFactory.mock_scan()
    # define the process_index is any tomwer_processes_existing
    if options.debug:
        _logger.setLevel(logging.DEBUG)

    window = SAAxisWindow()
    # window.setAttribute(qt.Qt.WA_DeleteOnClose)
    window.setWindowTitle("saaxis")
    window.setWindowIcon(icons.getQIcon("tomwer"))
    if scan.axis_params is None:
        scan.axis_params = QAxisRP()
    if scan.saaxis_params is None:
        scan.saaxis_params = QSAAxisParams()
    window.setScan(scan)
    splash.finish(window)
    window.show()
    qt.QApplication.restoreOverrideCursor()
    app.aboutToQuit.connect(window.stop)
    app.exec_()


def getinputinfo():
    return "tomwer saaxis [scanDir]"


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.quit()


if __name__ == "__main__":
    main(sys.argv)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import sys
from silx.gui import qt
import argparse
from tomwer.gui import icons
from tomwer.gui.utils.splashscreen import getMainSplashScreen
from tomwer.gui.reconstruction.ftserie import FtserieWidget
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.core.scan.scanfactory import ScanFactory
import signal

logging.basicConfig()
_logger = logging.getLogger(__name__)


def getinputinfo():
    return "tomwer ftseries [scan_path]"


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.quit()


def main(argv):
    import os

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "scan_path", help="Data file to show (h5 file, edf files, spec files)"
    )
    parser.add_argument(
        "--entry", default=None, help="An entry can be specify for hdf5 files"
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Set logging system in debug mode",
    )
    parser.add_argument(
        "--dry-run",
        dest="dry_run",
        action="store_true",
        default=False,
        help="Only create the .par file and print the command executed",
    )
    options = parser.parse_args(argv[1:])

    global app  # QApplication must be global to avoid seg fault on quit
    app = qt.QApplication.instance() or qt.QApplication([])

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler
    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catch
    timer.timeout.connect(lambda: None)

    splash = getMainSplashScreen()
    options.scan_path = options.scan_path.rstrip(os.path.sep)
    if options.entry is not None:
        valid_entries = HDF5TomoScan.get_valid_entries(options.scan_path)
        if options.entry not in valid_entries:
            raise ValueError(
                "entry {} is invalid. Does it exists ? Is the "
                "file NXTomo compliant ?. Valid entry are {}"
                "".format(options.entry, valid_entries)
            )
        scan = HDF5TomoScan(scan=options.scan_path, entry=options.entry)
    else:
        scan = ScanFactory.create_scan_object(scan_path=options.scan_path)
    scan.set_process_index_frm_tomwer_process_file()

    widget = FtserieWidget()
    widget.setScan(scan=scan)
    # for the application we run for the reconstruction to be finished
    # to give back hand to the user
    widget.reconsStack._forceSync = True

    widget.set_dry_run(options.dry_run)
    widget.setWindowTitle("FtSerie")
    widget.setWindowIcon(icons.getQIcon("tomwer"))
    splash.finish(widget)
    widget.show()
    app.exec_()


if __name__ == "__main__":
    main(sys.argv)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "16/11/2020"


from silx.gui import qt
from Orange.widgets import widget, gui
from tomwer.web.client import OWClient
from Orange.widgets.settings import Setting
from Orange.widgets.widget import Output, Input
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.process.control.volumesymlink import OutputType, VolumeSymbolicLink
from orangecontrib.tomwer.orange.managedprocess import SuperviseOW
from processview.core.manager import ProcessManager, DatasetState
from tomwer.utils import docstring
import logging

_logger = logging.getLogger(__name__)


class VolumeSymLinkOW(SuperviseOW, OWClient):
    """
    link volume reconstructed at proposal file if possible else under
    the given folder
    """

    name = "volume symbolic link"
    id = "orange.widgets.tomwer.control.VolumeSymLinkOW.VolumeSymLinkOW"
    description = (
        "create a symbolic link for the latest volume reconstructed on a "
        "given folder"
    )
    icon = "icons/volumesymlink.svg"
    priority = 200
    category = "esrfWidgets"
    keywords = [
        "hdf5",
        "tomwer",
        "folder",
        "volume",
        "symlink",
        "tomography",
        "symbolic link",
    ]

    want_main_area = True
    resizing_enabled = True
    compress_signal = False

    _scans = Setting(list())

    class Inputs:
        data_in = Input(name="data", type=TomwerScanBase, doc="scan to be process")

    class Outputs:
        data_out = Output(
            name="data", type=TomwerScanBase, doc="one scan to be process"
        )

    def __init__(self, parent=None):
        SuperviseOW.__init__(self, parent)
        OWClient.__init__(self)
        self._mainWidget = _OutputDir(parent=self)
        layout = gui.vBox(self.mainArea, self.name).layout()
        layout.addWidget(self._mainWidget)

    @Inputs.data_in
    def process(self, scan):
        if scan is None:
            return
        symlink_process = VolumeSymbolicLink()
        symlink_process.set_properties(
            {
                "output_type": self._mainWidget.getOutputType(),
                "output_folder": self._mainWidget.getOutputFolder(),
            }
        )
        ProcessManager().notify_dataset_state(
            dataset=self.scan,
            process=self,
            state=DatasetState.ON_GOING,
        )
        try:
            scan = symlink_process.process(scan=scan)
        except Exception as e:
            mess = "Failed to create symbolic link for {}".format(str(scan))
            _logger.processFailed(mess)
            ProcessManager().notify_dataset_state(
                dataset=self.scan, process=self, state=DatasetState.FAILED, details=mess
            )
        else:
            mess = "Create symbolic link for {}".format(str(scan))
            _logger.processSucceed(mess)
            ProcessManager().notify_dataset_state(
                dataset=self.scan,
                process=self,
                state=DatasetState.SUCCEED,
                details=mess,
            )
        self.Outputs.data_out.send(scan)

    def reprocess(self, dataset):
        self.process(dataset)


class _OutputDir(qt.QGroupBox):
    def __init__(self, parent):
        qt.QGroupBox.__init__(self, parent)
        self._button_grp = qt.QButtonGroup(self)
        self._button_grp.setExclusive(True)
        self.setLayout(qt.QGridLayout())
        self._oneLevelCB = qt.QRadioButton(OutputType.ONE_LEVEL_UPPER.value, self)
        self._button_grp.addButton(self._oneLevelCB)
        self.layout().addWidget(self._oneLevelCB, 1, 0, 1, 3)
        self._staticOpt = qt.QRadioButton("", self)
        self._button_grp.addButton(self._staticOpt)
        self.layout().addWidget(self._staticOpt, 2, 0, 1, 1)
        self._qleFolder = qt.QLineEdit("", self)
        self.layout().addWidget(self._qleFolder, 2, 1, 1, 1)
        self._selectPB = qt.QPushButton("select", self)
        self.layout().addWidget(self._selectPB, 2, 2, 1, 1)

        # connect signal / slot
        self._selectPB.released.connect(self._selectFolder)
        self._qleFolder.editingFinished.connect(self._activeStaticOpt)

        # set up
        self._oneLevelCB.setChecked(True)

    def _activeStaticOpt(self, *args, **kwargs):
        self._staticOpt.setChecked(True)

    def _selectFolder(self):
        dialog = qt.QFileDialog(self)
        dialog.setWindowTitle("Select destination folder for symbolic link")
        dialog.setModal(1)
        dialog.setFileMode(qt.QFileDialog.DirectoryOnly)

        if not dialog.exec_():
            dialog.close()
            return None

        dest_folder = dialog.selectedFiles()[0]
        self._qleFolder.setText(dest_folder)

    def getOutputType(self) -> OutputType:
        if self._oneLevelCB.isChecked():
            return OutputType.ONE_LEVEL_UPPER
        elif self._staticOpt.isChecked():
            return OutputType.STATIC
        else:
            raise TypeError("Not managed")

    def getOutputFolder(self):
        return self._qleFolder.text()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "01/12/2016"

from Orange.widgets import widget, gui
from silx.gui import qt
from tomwer.core.utils import logconfig
from tomwer.gui.control.datalist import GenericDataListDialog
from tomwer.core.process.control.scanlist import _ScanList
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.web.client import OWClient
from tomwer.core.scan.scanbase import TomwerScanBase
from Orange.widgets.settings import Setting
from Orange.widgets.widget import Output, Input
import logging

logger = logging.getLogger(__name__)


class DataListOW(widget.OWWidget, OWClient):
    name = "data list"
    id = "orange.widgets.tomwer.scanlist"
    description = "List path to reconstructions/scans. Those path will be send to the next box once validated."
    icon = "icons/scanlist.svg"
    priority = 50
    category = "esrfWidgets"
    keywords = ["tomography", "file", "tomwer", "folder"]

    want_main_area = True
    resizing_enabled = True
    compress_signal = False

    _scanIDs = Setting(list())

    assert len(_ScanList.outputs) == 1

    class Outputs:
        data_out = Output(
            name=_ScanList.outputs[0].name,
            type=_ScanList.outputs[0].type,
            doc=_ScanList.outputs[0].doc,
        )

    class Inputs:
        data_in = Input(name="data", type=TomwerScanBase)

    def __init__(self, parent=None):
        """A simple annuary which register all folder containing completed scan

        .. warning: the widget won't check for scan validity and will only
            emit the path to folders to the next widgets

        :param parent: the parent widget
        """
        widget.OWWidget.__init__(self, parent)
        OWClient.__init__(self)
        self.widget = GenericDataListDialog(parent=self)
        self.widget.sigUpdated.connect(self._updateSettings)
        self._loadSettings()
        layout = gui.vBox(self.mainArea, self.name).layout()
        layout.addWidget(self.widget)
        self.widget._sendButton.clicked.connect(self._sendList)

        # expose API
        self.n_scan = self.widget.n_scan

        # alias used for the 'simple workflow' for now
        self.start = self._sendList
        self.clear = self.widget.clear

    def _sendList(self):
        """Send a signal for each list to the next widget"""

        def send(scan_):
            mess = "sending one scan %s" % d
            logger.debug(
                mess,
                extra={
                    logconfig.DOC_TITLE: self.widget._scheme_title,
                    logconfig.SCAN_ID: str(scan_),
                },
            )
            self.Outputs.data_out.send(scan_)

        for d in self.widget.datalist.items:
            if d in self.widget.datalist._scanIDs:
                send(scan_=self.widget.datalist._scanIDs[d])
            else:
                logger.error("unable to retrieve scan {}".format(d))

    @Inputs.data_in
    def add(self, scan):
        if scan is None:
            return

        self.widget.add(scan)

    def _loadSettings(self):
        for scan in self._scanIDs:
            assert isinstance(scan, str)
            if "@" in scan:
                entry, file_path = scan.split("@")
                hdf5_tomo_scan = HDF5TomoScan(entry=entry, scan=file_path)
                self.add(hdf5_tomo_scan)
            else:
                self.add(scan)

    def _updateSettings(self):
        self._scanIDs = []
        for scan in self.widget.datalist.items:
            self._scanIDs.append(scan)


def main():
    app = qt.QApplication([])
    s = DataListOW()
    s.show()
    app.exec_()


if __name__ == "__main__":
    main()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "04/03/2019"


import os
import copy

try:
    from contextlib import AbstractContextManager
except ImportError:
    from tomwer.third_party.contextlib import AbstractContextManager
from silx.utils.enum import Enum as _Enum
from tomwer.core.scan.scanbase import _TomwerBaseDock
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.core.process.baseprocess import SingleProcess, _input_desc, _output_desc
from nabu.resources.dataset_analyzer import EDFDatasetAnalyzer, HDF5DatasetAnalyzer
from nabu.resources.dataset_analyzer import DatasetAnalyzer
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.utils import docstring
from tomwer.core.process.reconstruction.nabu.utils import check_sinogram_half
from typing import Union
from nabu.io.config import generate_nabu_configfile
from nabu import version as nabu_version
from silx.io.dictdump import h5todict
from . import utils
from . import settings as nabu_settings
from tomoscan.io import HDF5File
import logging
from tomwer.core.scan.edfscan import EDFTomoScan
import subprocess
from silx.io.utils import h5py_read_dataset

_logger = logging.getLogger(__name__)

try:
    from nabu.app.local_reconstruction import LocalReconstruction
except ImportError as e:
    _logger.error(e)
    has_nabu = False
else:
    has_nabu = True


def run_slices_reconstruction(
    scan: TomwerScanBase,
    config: dict,
    dry_run: bool = False,
    local: bool = True,
) -> tuple:
    """
    call nabu for a reconstruction on scan with the given configuration

    :param TomwerScanBase scan: scan to reconstruct
    :param dict config: configuration to run the reconstruction
    :param bool dry_run: do we want to run dry
    :param bool local: do we want to run a local reconstruction
    :param stderr: file to redirect stderr
    :param stdout: file to redirect stdout
    """
    check_sinogram_half(scan)
    _logger.info("start reconstruction of {}".format(str(scan)))
    # if scan contains some center of position copy it to nabu
    if scan.axis_params is not None and scan.axis_params.relative_cor_value is not None:
        if "reconstruction" in config:
            # move the cor value to the nabu reference
            cor_nabu_ref = scan.axis_params.relative_cor_value + scan.dim_1 // 2
            config["reconstruction"]["rotation_axis_position"] = str(cor_nabu_ref)

    _logger.info("set nabu reconstruction parameters to {}".format(str(scan)))
    scan.nabu_recons_params = config

    nabu_configurations = interpret_tomwer_configuration(config, scan=scan)
    db_list = utils.retrieve_lst_of_value_from_str(
        config["phase"]["delta_beta"], type_=float
    )
    ask_sinogram_registration = len(db_list) > 0
    ask_sinogram_load = False
    output_urls = []
    stderrs = []
    stdouts = []
    final_configs = []
    all_succeed = True
    for nabu_configuration in nabu_configurations:
        l_config, slice_index = nabu_configuration
        succeed, urls, stdout, stderr, f_config = run_single_slice_reconstruction(
            config=l_config,
            scan=scan,
            local=local,
            slice_index=slice_index,
            dry_run=dry_run,
            ask_sinogram_registration=ask_sinogram_registration,
            ask_sinogram_load=ask_sinogram_load,
        )
        output_urls.extend(urls)
        # if slice_index is None this mean that we are simply creating the
        # .cfg file for nabu full volume.
        if slice_index is not None:
            stderrs.append(stderr)
            stdouts.append(stdout)
            final_configs.append(f_config)
        all_succeed = all_succeed and succeed
        if ask_sinogram_registration:
            ask_sinogram_load = True
            ask_sinogram_registration = False
    # tag latest reconstructions
    scan.set_latest_reconstructions(output_urls)
    return all_succeed, stdouts, stderrs, final_configs


class NabuSlices(SingleProcess):
    """Definition of the nabu reconstruction Single process"""

    inputs = [
        _input_desc(
            name="change recons params",
            type=_TomwerBaseDock,
            handler="updateReconsParam",
            doc="input with scan + reconstruction parameters",
        ),
        _input_desc(
            name="data", type=TomwerScanBase, handler="pathReceived", doc="scan path"
        ),
    ]
    # Note : scanReady don't intend to find an 'octave_FT_params.h5' file at
    # the folder level.
    # But updateReconsParam should always have a .h5 file defined
    outputs = [_output_desc(name="data", type=TomwerScanBase, doc="scan path")]

    def __init__(self, *arg, **kwargs):
        SingleProcess.__init__(self, *arg, **kwargs)
        self._dry_run = False

    def process(self, scan=None):
        if scan is None:
            return None
        if isinstance(scan, TomwerScanBase):
            scan = scan
        elif isinstance(scan, dict):
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            raise ValueError(
                "input type of {}: {} is not managed" "".format(scan, type(scan))
            )

        run_slices_reconstruction(
            scan=scan, config=self.get_configuration(), dry_run=self.dry_run
        )
        # register result
        entry = "entry"
        if isinstance(scan, HDF5TomoScan):
            entry = scan.entry

        with scan.acquire_process_file_lock():
            self.register_process(
                process_file=scan.process_file,
                entry=entry,
                configuration=self.get_configuration(),
                results={},
                process_index=scan.pop_process_index(),
                overwrite=True,
            )
        return scan

    @docstring(SingleProcess)
    def set_properties(self, properties):
        self.set_configuration(configuration=properties["_rpSetting"])

    def set_configuration(self, configuration: dict) -> None:
        SingleProcess.set_configuration(self, configuration=configuration)
        if "dry_run" in configuration:
            self.set_dry_run(bool(configuration["dry_run"]))

    def pathReceived(self, scan):
        return self.process(scan=scan)

    @staticmethod
    def program_name():
        return "nabu-slices"

    @staticmethod
    def program_version():
        return nabu_version

    def set_dry_run(self, dry_run):
        self._dry_run = dry_run

    @property
    def dry_run(self):
        return self._dry_run

    @staticmethod
    def get_process_frm_process_file(process_file, entry):
        """
        Read informations regarding the nabu process save in the
        tomwer_process.h5 file

        :param process_file:
        :param entry:
        :return: dictionary with the contain of the nabu process
        :rtype:dict
        """
        if entry is None:
            with HDF5File(process_file, "r", swmr=True) as h5f:
                entries = NabuSlices._get_process_nodes(
                    root_node=h5f, process=NabuSlices
                )
                if len(entries) == 0:
                    _logger.info("unable to find a Axis process in %s" % process_file)
                    return None
                elif len(entries) > 1:
                    raise ValueError("several entry found, entry should be " "specify")
                else:
                    entry = list(entries.keys())[0]
                    _logger.info("take %s as default entry" % entry)

        configuration_path = None
        res = {}

        with HDF5File(process_file, "r", swmr=True) as h5f:
            nabu_nodes = NabuSlices._get_process_nodes(
                root_node=h5f[entry], process=NabuSlices
            )
            index_to_path = {}
            for key, index in nabu_nodes.items():
                index_to_path[index] = key

            if len(nabu_nodes) == 0:
                return None
            # take the last processed dark ref
            last_process_index = sorted(list(nabu_nodes.values()))[-1]
            last_process_dark = index_to_path[last_process_index]
            if (len(index_to_path)) > 1:
                _logger.debug(
                    "several processing found for dark-ref, "
                    "take the last one: %s" % last_process_dark
                )

            for key_name in (
                "class_instance",
                "date",
                "program",
                "sequence_index",
                "version",
            ):
                if key_name in h5f[last_process_dark]:
                    res[key_name] = h5py_read_dataset(h5f[last_process_dark][key_name])
            if "configuration" in h5f[last_process_dark]:
                configuration_path = "/".join(
                    (h5f[last_process_dark].name, "configuration")
                )

        if configuration_path is not None:
            res["configuration"] = h5todict(
                h5file=process_file, path=configuration_path
            )
        return res


def interpret_tomwer_configuration(
    config: dict, scan: Union[TomwerScanBase, None]
) -> tuple:
    """
    tomwer can 'mock' the nabu reconstruction to request more feature.
    Typical use case is that we can ask for reconstruction of several
    slices and not only the volume

    :param dict config: tomwer configuration for nabu
    :param scan: requested if we want to get slices
    :return: tuple of tuples (nabu configuration, is slice)
    """

    def get_nabu_config(config):
        nabu_config = copy.deepcopy(config)
        if "tomwer_slices" in nabu_config:
            del nabu_config["tomwer_slices"]
        return nabu_config

    if "tomwer_slices" in config and scan is not None:
        slices = list(NabuSliceMode.getSlices(config["tomwer_slices"], scan))
    else:
        slices = []

    if "phase" in config and "delta_beta" in config["phase"]:
        pag_dbs = utils.retrieve_lst_of_value_from_str(
            config["phase"]["delta_beta"], type_=float
        )
        if len(pag_dbs) == 0:
            pag_dbs = (None,)
    else:
        pag_dbs = (None,)

    # by default add the slice 'None' which is the slice for the volume
    slices.append(None)
    nabu_config = get_nabu_config(config=config)
    res = []
    for slice_ in slices:
        for pag_db in pag_dbs:
            local_config = copy.deepcopy(nabu_config)
            if slice_ is not None:
                local_config["reconstruction"]["start_z"] = slice_
                local_config["reconstruction"]["end_z"] = slice_
            if pag_db is not None:
                local_config["phase"]["delta_beta"] = str(pag_db)
            res.append((local_config, slice_))
    return tuple(res)


def _get_file_basename_reconstruction(scan, slice_index, pag, db):
    """

    :param TomwerScanBase scan: scan reconstructed
    :param Union[None, int] slice_index: index of the slice reconstructed.
                                         if None, we want to reconstruct the
                                         entire volume
    :param bool pag: is it a paganin reconstruction
    :param int db: delta / beta parameter
    :return: basename of the file reconstructed (without any extension)
    """
    assert type(db) in (int, type(None))
    if isinstance(scan, HDF5TomoScan):
        basename, _ = os.path.splitext(scan.master_file)
        basename = os.path.basename(basename)
    else:
        basename = os.path.basename(scan.path)
    if slice_index is None:
        if pag:
            return "_".join((basename + "pag", "db" + str(db).zfill(4)))
        else:
            return basename
    else:
        if pag:
            return "_".join(
                (
                    basename + "slice_pag",
                    str(slice_index).zfill(4),
                    "db" + str(db).zfill(4),
                )
            )
        else:
            return "_".join((basename + "slice", str(slice_index).zfill(4)))


def run_single_slice_reconstruction(
    scan,
    config,
    dry_run,
    slice_index,
    local,
    ask_sinogram_registration: bool,
    ask_sinogram_load: bool,
) -> tuple:
    """
    # TODO: might need something like a context or an option "keep" slice in memory

    :param scan:
    :param config:
    :param dry_run:
    :param Union[None,int] slice_index: slice index to reconstruct
    :param local:
    :param stdout: file to redirect stdout
    :param stderr: file to redirect stderr
    :param bool ask_sinogram_registration: should we ask nabu to register sinogram
    :param bool ask_sinogram_load: should we ask nabu to load sinogram
    :return: (list of output urls, process.stdout, process.stderr)
    """
    dataset_params = scan.get_nabu_dataset_info(scan)
    if "dataset" in config:
        dataset_params.update(config["dataset"])
    config["dataset"] = dataset_params

    if local is True:
        resources_method = "local"
    else:
        resources_method = "slurm"
    config["resources"] = utils.get_nabu_resources_desc(
        scan=scan, workers=1, method=resources_method
    )
    # force overwrite results
    if "output" not in config:
        config["output"] = {}
    # handle nabu sinogram step
    if scan.process_file is not None:
        steps_file_basename, _ = os.path.splitext(scan.process_file)
        steps_file_basename = "_".join(
            ("steps_file_basename", "nabu", "sinogram", "save", "step")
        )
        steps_file_basename = steps_file_basename + ".hdf5"
        steps_file = os.path.join(
            os.path.dirname(scan.process_file), steps_file_basename
        )
    else:
        steps_file = ""

    config["output"].update({"overwrite_results": 1})
    config["pipeline"] = {
        "save_steps": "sinogram" if ask_sinogram_registration else "",
        "resume_from_step": "sinogram" if ask_sinogram_load else "",
        "steps_file": steps_file,
    }

    def treateOutputConfig(_config):
        """
        - add or overwrite some parameters of the dictionary
        - create the output directory if does not exist
        """
        pag = False
        db = None
        if "phase" in _config:
            if "method" in _config["phase"] and _config["phase"]["method"] != "":
                pag = True
                if "delta_beta" in _config["phase"]:
                    db = round(float(_config["phase"]["delta_beta"]))
        if "output" in config:
            _file_name = _get_file_basename_reconstruction(
                scan=scan, slice_index=slice_index, pag=pag, db=db
            )
            _config["output"]["file_prefix"] = _file_name
            if _config["output"]["location"] not in ("", None):
                # if user specify the location
                if not os.path.isdir(_config["output"]["location"]):
                    os.makedirs(_config["output"]["location"])
            else:
                # otherwise default location will be the data root level
                _config["output"]["location"] = scan.path
        return _config

    config = treateOutputConfig(config)
    # the policy is to save nabu .cfg file at the same location as the
    # force overwrite results

    cfg_folder = os.path.join(
        config["output"]["location"], nabu_settings.NABU_CFG_FILE_FOLDER
    )
    if not os.path.exists(cfg_folder):
        os.makedirs(cfg_folder)

    name = config["output"]["file_prefix"] + nabu_settings.NABU_CONFIG_FILE_EXTENSION
    if not isinstance(scan, EDFTomoScan):
        name = "_".join((scan.entry, name))
    conf_file = os.path.join(cfg_folder, name)
    _logger.info("{}: create {}".format(scan, conf_file))

    # add some tomwer metadata and save the configuration
    # note: for now the section is ignored by nabu but shouldn't stay that way
    with utils.TomwerInfo(config) as config_to_dump:
        generate_nabu_configfile(
            fname=conf_file, config=config_to_dump, options_level="advanced"
        )

    if slice_index is not None and dry_run is False and local:
        if not has_nabu:
            raise ImportError("Fail to import nabu")
        _logger.info(
            "run nabu slice reconstruction for %s with %s" "" % (scan.path, config)
        )

        file_format = config_to_dump["output"]["file_format"]
        # need to be executed in his own context
        command = " ".join(
            ("python", "-m", "nabu.resources.cli.reconstruct", conf_file)
        )
        _logger.info('call nabu from "{}"'.format(command))

        process = subprocess.Popen(
            command,
            shell=True,
            cwd=scan.path,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        outs, errs = process.communicate()

        recons_urls = utils.get_recons_urls(
            file_prefix=config_to_dump["output"]["file_prefix"],
            location=config_to_dump["output"]["location"],
            slice_index=None,
            scan=scan,
            file_format=file_format,
            start_z=None,
            end_z=None,
        )

        return (
            len(recons_urls) > 0,
            recons_urls,
            outs,
            errs,
            config_to_dump,
        )
    else:
        return (True, [], None, None, None)


def get_nabu_dataset_analyzer(scan: TomwerScanBase) -> DatasetAnalyzer:
    """

    :param scan:
    :return:
    """
    if isinstance(scan, EDFTomoScan):
        return EDFDatasetAnalyzer(location=scan.path)
    elif isinstance(scan, HDF5TomoScan):
        assert os.path.exists(scan.master_file)
        assert os.path.isfile(scan.master_file)
        return HDF5DatasetAnalyzer(location=scan.master_file)
    else:
        raise TypeError("given scan type %s is not managed" % type(scan))


class NabuSliceMode(_Enum):
    MIDDLE = "middle"
    OTHER = "other"

    @staticmethod
    def getSlices(slices, scan) -> tuple:
        res = []
        try:
            mode = NabuSliceMode.from_value(slices)
        except ValueError:
            try:
                res = utils.retrieve_lst_of_value_from_str(slices, type_=int)
            except Exception:
                pass
        else:
            if mode == mode.MIDDLE:
                n_slice = scan.dim_2 or 2048
                res.append(n_slice // 2)
            else:
                raise ValueError(
                    "there should be only two ways of defining "
                    "slices: middle one or other, by giving "
                    "an unique value or a list or a tuple"
                )
        return tuple(res)
